/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/*
 * includes
 */

#include "std.h"
#include "inject.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <commdlg.h>
#include <direct.h>
#include "resource.h"
#include "version.h"
#include "state.h"

/*
 * defines
 */

#define MAX_GAMES		1024

#define KEY_JOY			0x01
#define KEY_KBD			0x02

#define HFLAGS_RINC_SHIFT	0x00
#define HFLAGS_RINC_MASK	0x7f
#define HFLAGS_SCR_SHIFT	0x07
#define HFLAGS_SCR_MASK		0x03

/*
 * typedefs
 */

/*
 * structs
 */

class Header
{
public:
	char	sig[8];
	char	name[8];
	int32	len;
	uint32	flags;
	KeyMap	keys[16];
};

class KeyDef
{
public:
	char	*name;
	KeyMap	m;
};

class CGameCache;
class CGame;

class CGameCacheEntry
{
public:
	CGameCache	*owner;
	CGameCacheEntry	*prev,*next;
	char		*filename;
	Header		h;
	bool		defined;

	CGameCacheEntry(CGameCache *own);
	~CGameCacheEntry();
};

class CGameCache
{
public:
	CGameCacheEntry	*e;

	CGameCache(void);
	~CGameCache();

	void			Load(void);
	void			Save(void);
	CGameCacheEntry*	Find(CGame *g);
};


class CGame
{
public:
	char		*fileName;
	byte		keyIds[16];
//	Header		h;

	CGameCacheEntry	*gc;
};

/*
 * globals
 */

char			initialDir[1024];

int32			currentGameId=-1;
HINSTANCE		hInstance;
HWND			hwndMain;
HWND			hwndProp;
char			tmp[1024];
char			tmp2[1024];

int32			gameCt=0;
CGame			games[MAX_GAMES];

byte			mem[65536];
char*			sig="STRAGS!!";

KeyDef		keyDefs[]=
{
	{	"Kmpstn Fire2",		{	KEY_JOY,	0x1f,		0x20,		0x00	}	},
	{	"Kmpstn Fire",		{	KEY_JOY,	0x1f,		0x10,		0x00	}	},
	{	"Kmpstn UP",		{	KEY_JOY,	0x1f,		0x08,		0x00	}	},
	{	"Kmpstn DOWN",		{	KEY_JOY,	0x1f,		0x04,		0x00	}	},
	{	"Kmpstn LEFT",		{	KEY_JOY,	0x1f,		0x02,		0x00	}	},
	{	"Kmpstn RIGHT",		{	KEY_JOY,	0x1f,		0x01,		0x00	}	},

	{	"SHIFT",	{	KEY_KBD,	0x00,		0x01,		0x00	}	},
	{	"Z",		{	KEY_KBD,	0x00,		0x02,		0x00	}	},
	{	"X",		{	KEY_KBD,	0x00,		0x04,		0x00	}	},
	{	"C",		{	KEY_KBD,	0x00,		0x08,		0x00	}	},
	{	"V",		{	KEY_KBD,	0x00,		0x10,		0x00	}	},
	{	"A",		{	KEY_KBD,	0x01,		0x01,		0x00	}	},
	{	"S",		{	KEY_KBD,	0x01,		0x02,		0x00	}	},
	{	"D",		{	KEY_KBD,	0x01,		0x04,		0x00	}	},
	{	"F",		{	KEY_KBD,	0x01,		0x08,		0x00	}	},
	{	"G",		{	KEY_KBD,	0x01,		0x10,		0x00	}	},
	{	"Q",		{	KEY_KBD,	0x02,		0x01,		0x00	}	},
	{	"W",		{	KEY_KBD,	0x02,		0x02,		0x00	}	},
	{	"E",		{	KEY_KBD,	0x02,		0x04,		0x00	}	},
	{	"R",		{	KEY_KBD,	0x02,		0x08,		0x00	}	},
	{	"T",		{	KEY_KBD,	0x02,		0x10,		0x00	}	},
	{	"1",		{	KEY_KBD,	0x03,		0x01,		0x00	}	},
	{	"2",		{	KEY_KBD,	0x03,		0x02,		0x00	}	},
	{	"3",		{	KEY_KBD,	0x03,		0x04,		0x00	}	},
	{	"4",		{	KEY_KBD,	0x03,		0x08,		0x00	}	},
	{	"5",		{	KEY_KBD,	0x03,		0x10,		0x00	}	},
	{	"0",		{	KEY_KBD,	0x04,		0x01,		0x00	}	},
	{	"9",		{	KEY_KBD,	0x04,		0x02,		0x00	}	},
	{	"8",		{	KEY_KBD,	0x04,		0x04,		0x00	}	},
	{	"7",		{	KEY_KBD,	0x04,		0x08,		0x00	}	},
	{	"6",		{	KEY_KBD,	0x04,		0x10,		0x00	}	},
	{	"P",		{	KEY_KBD,	0x05,		0x01,		0x00	}	},
	{	"O",		{	KEY_KBD,	0x05,		0x02,		0x00	}	},
	{	"I",		{	KEY_KBD,	0x05,		0x04,		0x00	}	},
	{	"U",		{	KEY_KBD,	0x05,		0x08,		0x00	}	},
	{	"Y",		{	KEY_KBD,	0x05,		0x10,		0x00	}	},
	{	"ENTER",	{	KEY_KBD,	0x06,		0x01,		0x00	}	},
	{	"L",		{	KEY_KBD,	0x06,		0x02,		0x00	}	},
	{	"K",		{	KEY_KBD,	0x06,		0x04,		0x00	}	},
	{	"J",		{	KEY_KBD,	0x06,		0x08,		0x00	}	},
	{	"H",		{	KEY_KBD,	0x06,		0x10,		0x00	}	},
	{	"SPACE",	{	KEY_KBD,	0x07,		0x01,		0x00	}	},
	{	"SYMSHIFT",	{	KEY_KBD,	0x07,		0x02,		0x00	}	},
	{	"M",		{	KEY_KBD,	0x07,		0x04,		0x00	}	},
	{	"N",		{	KEY_KBD,	0x07,		0x08,		0x00	}	},
	{	"B",		{	KEY_KBD,	0x07,		0x10,		0x00	}	},
	{	NULL,		{	0,		0,		0,		0	}	},
};


char *defaultKeys[]=
{
	"Kmpstn UP",
	"Kmpstn DOWN",
	"Kmpstn LEFT",
	"Kmpstn RIGHT",
	"SPACE",
	"Kmpstn Fire",
	"ENTER",
	"0",
	"1",
};

int32 keyBits[]=
{
	15-6,15-7,15-5,15-4,15-2,15-0,15-1,15-9,15-8
};
int32 bitsValid[16];

OPENFILENAME		ofn;
static char szFilter[]=
"GBA Binaries (*.BIN)\0*.res\0" \
"All files (*.*)\0*.*\0\0";

char	binFilename[1024];
char	binTitlename[1024];

CGameCache		gc;

HBITMAP		hDib;
void*		bits=NULL;
BITMAPINFO	*bmi;

/*
 * function prototypes
 */

BOOL CALLBACK WINAPI	MainDlgProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);
BOOL CALLBACK WINAPI	PropDlgProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);
void			ProcessFiles(HDROP h);
int32			VerifyZ80(char *fn);
int32			CreateImage(char *fn);
void			UpdateHeaderKeys(CGame *g);
void			BmpInit(HWND hwnd);
LRESULT CALLBACK	ScreenProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);
void			FetchScreen(CGame *g);

/*
 * code
 */


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int WINAPI WinMain(HINSTANCE hInst,HINSTANCE hPrevInstance,LPSTR cmd,int show)
{
	int32		i;

//	int32 LoadZ80(CState *s,char *fn);
//	CState		s;
//	for(i=0;i<MEMBLK_CT;i++) s.SetBlockPtr(i,new byte[16384]);
//	LoadZ80(&s,"trapdoor.z80");
//	i=s.Save(NULL,0L);


	hInstance=hInst;

	if(!hPrevInstance)
	{
		WNDCLASSEX	wc;
		memset(&wc,0,sizeof(wc));
		wc.cbSize=sizeof(wc);
		wc.lpfnWndProc=ScreenProc;
		wc.hInstance=hInstance;
		wc.lpszClassName="SCREENSHOT";
		RegisterClassEx(&wc);
	}

	memset(bitsValid,0,sizeof(bitsValid));
	for(i=0;i<9;i++)
	{
		bitsValid[ keyBits[i] ]=1;
	}

	getcwd(initialDir,sizeof(initialDir));
	gc.Load();

	DialogBox(hInstance,MAKEINTRESOURCE(IDD_MAIN),NULL,MainDlgProc);

	for(i=0;i<gameCt;i++)	UpdateHeaderKeys(&games[i]);

	chdir(initialDir);
	gc.Save();

	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

BOOL CALLBACK WINAPI MainDlgProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	HDROP		h;
	int32		i;
	char		tmp[64];

	switch(msg)
	{
	case WM_INITDIALOG:
		hwndMain=hwnd;
		DragAcceptFiles(hwnd,TRUE);

		sprintf(tmp,"FooN Injector %s",FOON_VERSION_STRING);
		SetWindowText(hwnd,tmp);
		break;
	case WM_CLOSE:
		EndDialog(hwnd,0);
		return(TRUE);
		break;
	case WM_DROPFILES:
		h=(HDROP)wParam;
		ProcessFiles(h);
		break;
	case WM_VKEYTOITEM:
		if(LOWORD(wParam)==VK_DELETE)
		{
			if(lParam==(LPARAM)GetDlgItem(hwnd,IDC_LIST))
			{
				i=SendDlgItemMessage(hwnd,IDC_LIST,LB_GETCURSEL,0,0);
				if(i!=LB_ERR)
				{
					SendDlgItemMessage(hwnd,IDC_LIST,LB_DELETESTRING,i,0);
					if(games[i].fileName) free(games[i].fileName);
					memcpy(&games[i],&games[i+1],(gameCt-i)*sizeof(CGame));
					gameCt--;
				}
			}
		}
		return(-1);
		break;
	case WM_COMMAND:
		if(LOWORD(wParam)==IDC_LIST)
		{
			if(HIWORD(wParam)==LBN_DBLCLK)
			{
				currentGameId=SendDlgItemMessage(hwnd,IDC_LIST,LB_GETCURSEL,0L,0L);
				EnableWindow(hwnd,FALSE);
				DialogBox(hInstance,MAKEINTRESOURCE(IDD_PROP),NULL,PropDlgProc);
				EnableWindow(hwnd,TRUE);
			}
		}
		else if(LOWORD(wParam)==IDC_CREATE)
		{
			if(!gameCt)
			{
				MessageBox(hwndMain,"No games specified!","Error",MB_OK);
				return(FALSE);
			}
			GetDlgItemText(hwnd,IDC_OUTPUT,tmp,sizeof(tmp));
			CreateImage(tmp);
		}
		else if(LOWORD(wParam)==IDC_BROWSE)
		{
			OPENFILENAME		ofn;

			memset(&ofn,0,sizeof(ofn));
			ofn.lStructSize=sizeof(OPENFILENAME);
			ofn.hInstance=NULL;
			ofn.lpstrFilter=szFilter;
			ofn.lpstrCustomFilter=NULL;
			ofn.nMaxCustFilter=0;
			ofn.nFilterIndex=0;
			ofn.lpstrFile=NULL;
			ofn.nMaxFile=_MAX_PATH;
			ofn.lpstrFileTitle=NULL;
			ofn.nMaxFileTitle=_MAX_FNAME + _MAX_EXT;
			ofn.lpstrInitialDir=NULL;
			ofn.lpstrTitle=NULL;
			ofn.Flags=0;
			ofn.nFileOffset=0;
			ofn.nFileExtension=0;
			ofn.lpstrDefExt="bin";
			ofn.lCustData=0L;
			ofn.lpfnHook=NULL;
			ofn.lpTemplateName=NULL;

			ofn.hwndOwner=hwnd;
			ofn.lpstrFile=binFilename;
			ofn.lpstrFileTitle=binTitlename;
			ofn.Flags = OFN_OVERWRITEPROMPT;
			ofn.lpstrFilter=szFilter;
		
			if(!GetOpenFileName(&ofn))	break;

			SetDlgItemText(hwnd,IDC_OUTPUT,binFilename);
		}
		break;
	}
	return(FALSE);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void ProcessFiles(HDROP h)
{
	int32		fileCt;
	int32		i,j,k;
	CGameCacheEntry	*ge;

	// get file count
	fileCt=DragQueryFile(h,0xFFFFFFFF,NULL,0);

	// add files
	for(i=0;i<fileCt;i++)
	{
		DragQueryFile(h,i,tmp,sizeof(tmp));

		// got this already?
		if(SendDlgItemMessage(hwndMain,IDC_LIST,LB_FINDSTRINGEXACT,0L,(LPARAM)tmp)!=LB_ERR)
		{
			continue;
		}

		// how many?
		if(gameCt>=MAX_GAMES)
		{
			MessageBox(hwndMain,"You have reached the maximum number of games!","Error",MB_OK);
			break;
		}

		// verify file validity?
		if(VerifyZ80(tmp))
		{
			sprintf(tmp2,"%s is not a valid 48K .Z80 image",tmp);
			MessageBox(hwndMain,tmp2,"Error",MB_OK);
			continue;
		}

		// add game
		memset(&games[gameCt],0,sizeof(CGame));
		games[gameCt].fileName=strdup(tmp);
		ge=gc.Find(&games[gameCt]);

		for(k=strlen(tmp)-1;k>=0;k--)	if(tmp[k]=='\\') { k++; break; }
		memcpy(games[gameCt].gc->h.name,&tmp[k],8);
		for(j=0;j<8;j++) if(games[gameCt].gc->h.name[j]=='.') games[gameCt].gc->h.name[j]=0;


		if(!ge->defined)
		{
			// set default keys
			for(j=0;j<9;j++)
			{
				for(k=0;keyDefs[k].name;k++)
				{
					if(!stricmp(defaultKeys[j],keyDefs[k].name))
					{
						games[gameCt].keyIds[ keyBits[j] ]=(byte)k;
						break;
					}
				}
			}
			UpdateHeaderKeys(&games[gameCt]);

			// set default flags
			ge->h.flags=(3 << HFLAGS_SCR_SHIFT);

			ge->defined=true;
		}
		else
		{
			// it was defined - perform reverse mapping to keyids
			for(j=0;j<16;j++)
			{
				for(k=0;keyDefs[k].name;k++)
				{
					if(!memcmp(&games[gameCt].gc->h.keys[j],&keyDefs[k].m,sizeof(keyDefs[k].m)))
					{
						games[gameCt].keyIds[j]=(byte)k;
					}
				}
			}
		}


		SendDlgItemMessage(hwndMain,IDC_LIST,LB_INSERTSTRING,-1L,(LPARAM)tmp);
		SendDlgItemMessage(hwndMain,IDC_LIST,LB_SETITEMDATA,gameCt,(LPARAM)gameCt);

		gameCt++;
	}
	DragFinish(h);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 VerifyZ80(char *fn)
{
	FILE		*stream;
	int32		i,ver,flg;
	ver=-1;

	stream=fopen(fn,"rb");
	if(!stream)	return(-1);

	fgetc(stream);
	fgetc(stream);
	fgetc(stream);		fgetc(stream);
	fgetc(stream);		fgetc(stream);
	
	i=fgetc(stream);
	i+=fgetc(stream)<<8;
	if(i!=0)	{ ver=0; }

	fgetc(stream);	fgetc(stream);

	fgetc(stream);
	fgetc(stream);
	
	flg=fgetc(stream);

	fgetc(stream);
	fgetc(stream);
	fgetc(stream);		fgetc(stream);
	fgetc(stream);		fgetc(stream);
	fgetc(stream);		fgetc(stream);
	fgetc(stream);
	fgetc(stream);
	fgetc(stream);		fgetc(stream);
	fgetc(stream);		fgetc(stream);
	if(fgetc(stream))	{}
	if(fgetc(stream))	{}
	fgetc(stream);

	if(ver)
	{
		i=fgetc(stream);
		if(i==23)	ver=2;
		else if(i==54)	ver=3;
		else return(-3);

		i=fgetc(stream);	if(i!=0) { fclose(stream); return(-3); }

		fgetc(stream);		fgetc(stream);
		i=fgetc(stream);
		if(i!=0 && i!=1)
		{
			fclose(stream); return(-4);
		}
		fgetc(stream);
		i=fgetc(stream);	if(i==0xFF) { fclose(stream); return(-5); }
		fgetc(stream);
		fgetc(stream);
		i=ftell(stream);
		for(i=0;i<16;i++) fgetc(stream);

		if(ver==3) for(i=0;i<54-23;i++) fgetc(stream);
	}

	i=ftell(stream);

	while(1)
	{
		int32		pageLen,pageNum,pageOff;
		int32		wPtr,state,rept,val,len;

		if(feof(stream)) break;
		if(ver==0 && (flg&32))
		{
			pageOff=0x4000;
			i=ftell(stream); fseek(stream,0L,SEEK_END); pageLen=ftell(stream)-i; fseek(stream,i,SEEK_SET);
			if(!pageLen)	break;
		}
		else
		{
			pageLen=fgetc(stream);		pageLen+=fgetc(stream)<<8;
			if(feof(stream)) break;

			pageNum=fgetc(stream);
			if(pageNum==0)		pageOff=0x0000;
			else if(pageNum==4)	pageOff=0x8000;
			else if(pageNum==5)	pageOff=0xC000;
			else if(pageNum==8)	pageOff=0x4000;
			else
			{
				return(-6);
			}
		}

		wPtr=pageOff;
		len=0;
		state=0;
		while(len<pageLen)
		{
			i=fgetc(stream);	len++;

			if(state==0)
			{
				if(i!=0xED)	mem[wPtr++]=(byte)i;
				else		state=1;
			}
			else if(state==1)
			{
				if(i!=0xED)	{ mem[wPtr++]=(byte)0xED; mem[wPtr++]=(byte)i; state=0; }
				else		state=2;
			}
			else if(state==2)
			{
				rept=i;
				state=3;
			}
			else if(state==3)
			{
				val=i;
				while(rept--)	mem[wPtr++]=(byte)val;
				state=0;
			}
		}
	}

	fclose(stream);
	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

BOOL CALLBACK WINAPI PropDlgProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	int32		i,j,k;

	switch(msg)
	{
	case WM_INITDIALOG:
		hwndProp=hwnd;
		DragAcceptFiles(hwnd,TRUE);

		// populate combo boxes + set defaults
		for(i=0;i<9;i++)
		{
			for(j=0;keyDefs[j].name;j++)
			{
				SendDlgItemMessage(hwnd,IDC_UP+i,CB_INSERTSTRING,-1L,(LPARAM)keyDefs[j].name);
			}
			if(currentGameId!=-1)
			{
				SendDlgItemMessage(hwnd,IDC_UP+i,CB_SETCURSEL,games[currentGameId].keyIds[keyBits[i]],0);
			}
		}

		// populate R-INC
		for(i=0;i<8;i++)
		{
			char	tmp[16];
			sprintf(tmp,"%ld",i);
			SendDlgItemMessage(hwnd,IDC_RINC,CB_INSERTSTRING,-1L,(LPARAM)tmp);
		}
		SendDlgItemMessage(hwnd,IDC_RINC,CB_INSERTSTRING,-1L,(LPARAM)"Cycle Count/4");

		// populate Screen/Interrupt Timing
		SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_INSERTSTRING,-1L,(LPARAM)"Block 0");
		SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_INSERTSTRING,-1L,(LPARAM)"Block 1");
		SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_INSERTSTRING,-1L,(LPARAM)"Block 2");
		SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_INSERTSTRING,-1L,(LPARAM)"FullScreen");

		SendDlgItemMessage(hwnd,IDC_RINC,CB_SETCURSEL,0,0);
		SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_SETCURSEL,3,0);

		if(currentGameId!=-1)
		{
			uint32		f;
			f=games[currentGameId].gc->h.flags;
			SendDlgItemMessage(hwnd,IDC_RINC,CB_SETCURSEL,(f>>HFLAGS_RINC_SHIFT)&HFLAGS_RINC_MASK,0);
			SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_SETCURSEL,(f>>HFLAGS_SCR_SHIFT)&HFLAGS_SCR_MASK,0);
		}

		break;
	case WM_CLOSE:
		EndDialog(hwnd,0);
		return(TRUE);
		break;
	case WM_COMMAND:
		if(wParam==IDOK && currentGameId!=-1)
		{
			// update key ids
			for(i=0;i<9;i++)
			{
				j=SendDlgItemMessage(hwnd,IDC_UP+i,CB_GETCURSEL,0,0);
				games[currentGameId].keyIds[keyBits[i]]=(byte)j;
			}
			UpdateHeaderKeys(&games[currentGameId]);

			// get rinc/screenintmode
			j=SendDlgItemMessage(hwnd,IDC_RINC,CB_GETCURSEL,0,0);
			k=SendDlgItemMessage(hwnd,IDC_SCREENINTMODE,CB_GETCURSEL,0,0);
			games[currentGameId].gc->h.flags=((j&HFLAGS_RINC_MASK)<<HFLAGS_RINC_SHIFT) |
				                         ((k&HFLAGS_SCR_MASK)<<HFLAGS_SCR_SHIFT);

		}
		if(wParam==IDOK || wParam==IDCANCEL)
		{
			EndDialog(hwnd,0);
		}
		break;
	}
	return(FALSE);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 CreateImage(char *fn)
{
	FILE		*out;
	FILE		*stream;
	int32		i,j;

	out=fopen(fn,"wb");
	if(!out)	return(-1);

	// dump FooN image
	extern byte data[];
	extern long dataSize;
	fwrite(data,1,dataSize,out);

	// dump games
	for(i=0;i<gameCt;i++)
	{
		UpdateHeaderKeys(&games[i]);
	
		stream=fopen(games[i].fileName,"rb");
		if(!stream)
		{
			sprintf(tmp,"%s could not be read!");
			MessageBox(hwndMain,tmp,"Error",MB_OK);
			continue;
		}

		fseek(stream,0L,SEEK_END);
		games[i].gc->h.len=ftell(stream);
		fseek(stream,0L,SEEK_SET);
		memcpy(games[i].gc->h.sig,sig,sizeof(games[i].gc->h.sig));

		// write header
		fwrite(&games[i].gc->h,1,sizeof(games[i].gc->h),out);

		// write game
		while(1)
		{
			j=fread(tmp,1,sizeof(tmp),stream);
			if(j) fwrite(tmp,1,j,out);
			if(j!=sizeof(tmp)) break;
		}

		fclose(stream);

		// align!
		while(ftell(out)&7)	fputc(0x55,out);
	}

	// ensure there's no sig
	fputc(0,out);	fputc(0,out);	fputc(0,out);	fputc(0,out);
	fputc(0,out);	fputc(0,out);	fputc(0,out);	fputc(0,out);
	fclose(out);

	MessageBox(hwndMain,"Binary image created - have fun!","Image Created",MB_OK);

	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

CGameCache::CGameCache(void)
{
	e=NULL;
}

CGameCache::~CGameCache()
{
	while(e)	delete(e);
}

CGameCacheEntry::CGameCacheEntry(CGameCache *own)
{
	owner=own;

	// insert into list
	prev=NULL;	next=own->e;
	if(next)	next->prev=this;
	owner->e=this;

	filename=NULL;
	memset(&h,0,sizeof(h));

	defined=false;
}

CGameCacheEntry::~CGameCacheEntry()
{
	if(next)	next->prev=prev;
	if(prev)	prev->next=next;
	else		owner->e=next;
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

char *GetFileOnly(char *full)
{
	char		*c;

	c=&full[strlen(full)-1];
	while(c>full)
	{
		if((*c)=='\\' || (*c)=='/') { c++; break; }
		c--;
	}
	return(c);
}

void CGameCache::Load(void)
{
	FILE		*stream;
	int32		i,j;
	CGameCacheEntry	*c;

	// remove current entries
	while(e)	delete(e);

	stream=fopen("inject.ini","rb");
	if(!stream)	return;

	while(1)
	{
		j=fgetc(stream);
		if(feof(stream))	break;
		i=fgetc(stream);
		i|=fgetc(stream)<<8;

		c=new CGameCacheEntry(this);

		c->filename=new char[i+1];
		fread(c->filename,1,i,stream);
		c->filename[i]=0;

		fread(&c->h,1,sizeof(c->h),stream);
		c->defined=true;
	}

	fclose(stream);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void CGameCache::Save(void)
{
	FILE		*stream;
	CGameCacheEntry	*c;
	int32		i;

	stream=fopen("inject.ini","wb");
	if(!stream)	return;

	for(c=e;c!=NULL;c=c->next)
	{
		fputc(0,stream);
		i=strlen(c->filename);
		fputc(i&0xFF,stream);
		fputc(i>>8,stream);
		fwrite(c->filename,1,i,stream);
		fwrite(&c->h,1,sizeof(c->h),stream);
	}

	fclose(stream);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

CGameCacheEntry *CGameCache::Find(CGame *g)
{
	CGameCacheEntry		*c;
	char			*a;

	for(c=e;c!=NULL;c=c->next)
	{
		if(!stricmp(c->filename,g->fileName))
		{
			g->gc=c;
			return(c);
		}
	}

	// extract last part of filename
	a=GetFileOnly(g->fileName);
	for(c=e;c!=NULL;c=c->next)
	{
		if(!stricmp(GetFileOnly(c->filename),a))
		{
			g->gc=c;
			return(c);
		}
	}

	c=new CGameCacheEntry(this);
	c->filename=strdup(g->fileName);
	memset(&c->h,0,sizeof(c->h));
	g->gc=c;
	return(c);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void UpdateHeaderKeys(CGame *g)
{
	int32		j;

	for(j=0;j<16;j++)
	{
		if(bitsValid[j])
		{
			g->gc->h.keys[j]=keyDefs[ g->keyIds[j] ].m;
		}
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void BmpInit(HWND hwnd)
{
	int32		i;
	HDC		hdc;

	bmi=(BITMAPINFO*)malloc(sizeof(BITMAPINFO) + sizeof(RGBQUAD)*256);
	memset(bmi,0,sizeof(BITMAPINFO));
	bmi->bmiHeader.biSize=sizeof(bmi->bmiHeader);
	bmi->bmiHeader.biWidth=256;
	bmi->bmiHeader.biHeight=-192;
	bmi->bmiHeader.biBitCount=8;
	bmi->bmiHeader.biCompression=BI_RGB;
	bmi->bmiHeader.biPlanes=1;
	for(i=0;i<256;i++)
	{
		bmi->bmiColors[i].rgbBlue=0;
		bmi->bmiColors[i].rgbRed=0;
		bmi->bmiColors[i].rgbGreen=0;
		if(i&1) { bmi->bmiColors[i].rgbBlue|=0x80; if(i&8) bmi->bmiColors[i].rgbBlue|=0x7F; }
		if(i&2) { bmi->bmiColors[i].rgbRed|=0x80; if(i&8) bmi->bmiColors[i].rgbRed|=0x7F; }
		if(i&4) { bmi->bmiColors[i].rgbGreen|=0x80; if(i&8) bmi->bmiColors[i].rgbGreen|=0x7F; }
	}

	hdc=GetDC(hwnd);
	hDib=CreateDIBSection(hdc,bmi,DIB_RGB_COLORS,&bits,NULL,0L);
	ReleaseDC(hwnd,hdc);

	for(i=0;i<256*192;i++)
	{
		((byte*)bits)[i]=(byte)rand();
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

LRESULT CALLBACK ScreenProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	RECT		r;
	HDC		hdc;

	switch(msg)
	{
	case WM_CREATE:
		BmpInit(hwnd);
		memset(bits,0,256*192);
		if(currentGameId!=-1)	FetchScreen(&games[currentGameId]);
		break;
	case WM_PAINT:
		GetClientRect(hwnd,&r);
		hdc=GetDC(hwnd);
		FillRect(hdc,&r,(HBRUSH)GetStockObject(BLACK_BRUSH));
		SetDIBitsToDevice(hdc,(r.right-256)/2,(r.bottom-192)/2,256,192,0,0,0,192,bits,bmi,DIB_RGB_COLORS);
		ReleaseDC(hwnd,hdc);
		break;
	case WM_DESTROY:
		if(hDib)
		{
			DeleteObject(hDib);
			hDib=NULL;
		}
		break;
	}
	return(DefWindowProc(hwnd,msg,wParam,lParam));
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void FetchScreen(CGame *g)
{
	int32		i,k,l,m,n,o;

	memset(bits,0,256*192);
	i=VerifyZ80(g->fileName);
	if(i)	return;

	o=0;
	for(i=0;i<192;i++)
	{
		for(k=0;k<32;k++)
		{
			l=mem[((i/64)*2048) + (((i/8)&7)*32) + ((i&7)*256) + k + 16384];
			m=mem[(i/8)*32+k + 16384+6144];
			for(n=0;n<8;n++)
			{
				if(l & (1<<(7-n)))	((byte*)bits)[o]=((m>>0)&7) | ((m&64)?8:0);
				else			((byte*)bits)[o]=((m>>3)&7) | ((m&64)?8:0);
				o++;
			}
		}
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 LoadZ80(CState *s,char *fn)
{
	int32		i,j;
	int32		ver;
	int32		flg;
	FILE		*stream;

	stream=fopen(fn,"rb");
	if(!stream)	return(-1);

	// clear state
	s->Reset();

	ver=-1;

	s->SetReg(REG_A,fgetc(stream));		s->SetReg(REG_F,fgetc(stream));
	s->SetReg(REG_C,fgetc(stream));		s->SetReg(REG_B,fgetc(stream));
	s->SetReg(REG_L,fgetc(stream));		s->SetReg(REG_H,fgetc(stream));

	i=fgetc(stream);
	j=fgetc(stream);
	s->SetReg(REG_PCL,i);	s->SetReg(REG_PCH,j);
	if(i || j)	ver=0;

	s->SetReg(REG_SPL,fgetc(stream));	s->SetReg(REG_SPH,fgetc(stream));
	s->SetReg(REG_I,fgetc(stream));		s->SetReg(REG_R,fgetc(stream));
	flg=fgetc(stream);	s->SetReg(REG_BORDER,flg);

	s->SetReg(REG_E,fgetc(stream));		s->SetReg(REG_D,fgetc(stream));

	s->SetReg(REG_C2,fgetc(stream));	s->SetReg(REG_B2,fgetc(stream));
	s->SetReg(REG_E2,fgetc(stream));	s->SetReg(REG_D2,fgetc(stream));
	s->SetReg(REG_L2,fgetc(stream));	s->SetReg(REG_H2,fgetc(stream));
	s->SetReg(REG_A2,fgetc(stream));	s->SetReg(REG_F2,fgetc(stream));

	s->SetReg(REG_IYL,fgetc(stream));	s->SetReg(REG_IYH,fgetc(stream));
	s->SetReg(REG_IXL,fgetc(stream));	s->SetReg(REG_IXH,fgetc(stream));
	i=fgetc(stream);	j=fgetc(stream);
	i=(i?IFF_1:0) | (j?IFF_2:0);
	s->SetReg(REG_IFF,i);
	s->SetReg(REG_IM,fgetc(stream));

	// version dependent stuff...
	if(ver!=0)
	{
		// get extra header length
		i=fgetc(stream);	i+=fgetc(stream)<<8;
		if(i==23)	ver=2;
		else if(i==54)	ver=3;
		else
		{
			return(-3);
		}

		// get PC
		s->SetReg(REG_PCL,fgetc(stream));	s->SetReg(REG_PCH,fgetc(stream));

		// ensure it's not 128K!
		j=fgetc(stream);
		if(j!=0 && j!=1)	return(-4);

		fgetc(stream);
		j=fgetc(stream);	if(j==0xFF)	return(-5);

		fgetc(stream);	fgetc(stream);
		for(j=0;j<16;j++)	fgetc(stream);

		if(ver==3)	for(i=0;i<54-23;i++) fgetc(stream);
	}

	// read memory blocks
	while(1)
	{
		int32		pageLen,pageNum,realPage;
		int32		wPtr,state,rept,val,len;

		if(ver==0 && (flg&32))
		{
			realPage=MEMBLK_48K_4000;
			i=ftell(stream);
			fseek(stream,0L,SEEK_END); pageLen=ftell(stream)-i; fseek(stream,i,SEEK_SET);
		}
		else
		{
			pageLen=fgetc(stream);	pageLen+=fgetc(stream)<<8;
			if(feof(stream))	break;

			pageNum=fgetc(stream);
			if(pageNum==0)		realPage=MEMBLK_48K_ROM;
			else if(pageNum==4)	realPage=MEMBLK_48K_4000;
			else if(pageNum==5)	realPage=MEMBLK_48K_8000;
			else if(pageNum==8)	realPage=MEMBLK_48K_C000;
			else
			{
				return(-6);
			}
		}

		wPtr=0;
		len=0;
		state=0;
		rept=0;
		while(len<pageLen)
		{
			i=fgetc(stream);	len++;
			if(feof(stream))	break;

			if(state==0)
			{
				if(i!=0xED)	s->WriteMem(realPage,wPtr++,i);
				else		state=1;
			}
			else if(state==1)
			{
				if(i!=0xED)	{ s->WriteMem(realPage,wPtr++,0xED); s->WriteMem(realPage,wPtr++,i); state=0; }
				else		state=2;
			}
			else if(state==2)
			{
				rept=i;
				state=3;
			}
			else if(state==3)
			{
				val=i;
				while(rept--)
				{
					s->WriteMem(realPage,wPtr++,val);
					if(wPtr==16384) { wPtr=0; realPage++; }
				}
				state=0;
			}
			if(wPtr==16384) { wPtr=0; realPage++; }
		}
	}

	fclose(stream);
	return(0);
}


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

