/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/*
 * includes
 */

#include "std.h"
#include "state.h"

#include "specrom.cpp"

/*
 * defines
 */

/*
 * structures/classes
 */

/*
 * globals
 */

/*
 * prototypes
 */

/*
 * code
 */

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

CState::CState(void)
{
	memset(blocks,0,sizeof(blocks));
	Reset();
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::SetBlockPtr(int32 blockId,byte *data)
{
	if(blockId<0 || blockId>=MEMBLK_CT)	return;

	blocks[blockId]=data;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::WriteMem(int32 blockId,int32 offset,int32 val)
{
	if(blockId<0 || blockId>=MEMBLK_CT)	return;
	if(!blocks[blockId])			return;

	if(offset<0 || offset>65535)		return;
	blocks[blockId][offset]=(byte)val;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::SetButton(int32 butId,KeyMap k)
{
	keys[butId]=k;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::SetReg(int32 regId,int32 val)
{
	if(regId<0 || regId>=REG_CT)	return;
	regs[regId]=(byte)val;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::SetOpt(int32 optId,int32 val)
{
	if(optId<0 || optId>=OPT_CT)	return;
	opts[optId]=(byte)val;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::Reset(void)
{
	int32		i;

	memset(keys,0,sizeof(keys));
	memset(regs,0,sizeof(regs));
	memset(opts,0,sizeof(opts));
	strcpy(title,"unnamed");

	for(i=0;i<MEMBLK_CT;i++)
	{
		if(!blocks[i])	continue;
		memset(blocks[i],0,16384);
	}

	if(blocks[MEMBLK_48K_ROM])	memcpy(blocks[MEMBLK_48K_ROM],specRom,16384);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::SetTitle(char *t)
{
	strcpy(title,t);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::AllocBlockPtrs(void)
{
	int32		i;

	for(i=0;i<MEMBLK_CT;i++)
	{
		blocks[i]=new byte[16384];
		memset(blocks[i],0,16384);
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::EmitByte(byte c)
{
	if(destData && destOff<destSize)	destData[destOff]=c;
	destOff++;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::EmitWord(uint16 c)
{
	EmitByte((byte)(c>>8));
	EmitByte((byte)(c&0xFF));
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::EmitDword(uint32 c)
{
	EmitWord((uint16)(c>>16));
	EmitWord((uint16)(c&0xFFFF));
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

int32 CState::Save(byte *data,int32 maxLen)
{
	int32		headerPt;
	int32		i,j;
	int32		bp;

	// prepare write values
	destData=data;
	destSize=maxLen;
	destOff=0;

	// emit header block
	EmitDword(SIG_HEADER);	headerPt=destOff;	EmitDword(0x00);

	// emit title
	EmitDword(SIG_TITLE);	EmitDword(strlen(title)+1);
	for(i=0;i<(int32)strlen(title)+1;i++)	EmitByte((byte)(title[i]));

	// emit regs
	EmitDword(SIG_REGS);	EmitDword(sizeof(regs));
	for(i=0;i<(int32)sizeof(regs);i++)	EmitByte(regs[i]);

	// emit opts
	EmitDword(SIG_OPTS);	EmitDword(sizeof(opts));
	for(i=0;i<(int32)sizeof(opts);i++)	EmitByte(opts[i]);

	// emit keys
	EmitDword(SIG_KEYS);	EmitDword(sizeof(keys));
	for(i=0;i<sizeof(keys)/sizeof(KeyMap);i++)
	{
		EmitByte(keys[i].type);
		EmitByte(keys[i].port);
		EmitByte(keys[i].val);
		EmitByte(keys[i].pad);
	}

	// emit memory
	for(i=0;i<MEMBLK_CT;i++)
	{
		if(!blocks[i])	continue;

		// block non-default?
		if(i==MEMBLK_48K_ROM)
		{
			if(!memcmp(specRom,blocks[i],16384))	continue;
		}
		else if(i==MEMBLK_128K_ROM)
		{
//			if(!memcmp(specRom,blocks[i],16384))	continue;
			continue;
		}
		else
		{
			for(j=0;j<16384;j++)	if(blocks[i][j])	break;
			if(j==16384)	continue;
		}

		// emit block
		EmitDword(SIG_MEMBLK_BASE+i);
		bp=destOff;
		EmitDword(0);
		EmitBlock(blocks[i],16384);
		j=destOff;	destOff=bp;
		EmitDword((j-bp)-4);
		destOff=j;
	}

	// fill in header length field
	i=destOff;
	destOff=headerPt;
	EmitDword((i-headerPt)-4);
	destOff=i;

	return(destOff);

	return(0);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::EmitBlock(byte *d,int32 len)
{
	int32		i;
	int32		lastCt,l;
	byte		lastByte;

	lastByte=-1;
	lastCt=0;

	// simple RLE compression:
	// FF FF = FF
	// FF CT VAL = VAL x CT+1
	for(i=0;i<16385;i++)
	{
		if(i==16384 || d[i]!=lastByte)
		{
			// emit lastCt x lastByte
			if(lastCt==1)
			{
				if(lastByte==0xFF)	EmitByte(0xFF);
				EmitByte(lastByte);
			}
			else if(lastCt==2)
			{
				if(lastByte==0xFF)	{ EmitByte(0xFF); EmitByte(0x00); EmitByte(0xFF); }
				else			{ EmitByte(lastByte); EmitByte(lastByte); }
			}
			else
			{
				while(lastCt)
				{
					l=lastCt;
					if(l>256)	l=256;
					EmitByte(l-2);
					EmitByte(lastByte);
					lastCt-=l;
				}
			}

			// new vals
			lastByte=d[i];
			lastCt=1;
		}
		else
		{
			lastCt++;
		}
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

byte CState::ReadByte(void)
{
	if(srcOff >= srcSize)
	{
		srcOverflow=true;
		return(0);
	}
	return(srcData[srcOff++]);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

uint16 CState::ReadWord(void)
{
	uint16		q;
	q=ReadByte();	q<<=8;
	q|=ReadByte();
	return(q);
}

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

uint32 CState::ReadDword(void)
{
	uint32		q;
	q=ReadWord();	q<<=16;
	q|=ReadWord();
	return(q);
}

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void CState::ReadBlock(void)
{
	int32		i,j;

	while(srcOff < srcSize)
	{
		i=ReadByte();
		if(i==0xFF)
		{
			i=ReadByte();
			if(i==0xFF)	EmitByte((byte)i);
			else
			{
				j=ReadByte();
				i+=2;
				while(i--)	EmitByte((byte)j);
			}
		}
		else
		{
			EmitByte((byte)i);
		}
	}
}

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

int32 CState::Load(byte *data,int32 maxLen)
{
	int32		i,j;
	uint32		d,e,f;

	srcOff=0;
	srcData=data;
	srcSize=maxLen;

	// read main block + len
	d=ReadDword();
	if(d!=SIG_HEADER)	return(-1);
	d=ReadDword();
	if(d+8 > (uint32)maxLen)	return(-1);

	while(1)
	{
		j=srcOff;

		// get sig and len
		e=ReadDword();
		f=ReadDword();

		switch(e)
		{
		case SIG_TITLE:
			for(i=0;i<(int32)f;i++)	title[i]=ReadByte();
			break;
		case SIG_KEYS:
			for(i=0;i<sizeof(keys)/sizeof(KeyMap);i++)
			{
				keys[i].type=ReadByte();
				keys[i].port=ReadByte();
				keys[i].val=ReadByte();
				keys[i].pad=ReadByte();
			}
			break;
		case SIG_OPTS:
			for(i=0;i<sizeof(opts);i++)	opts[i]=ReadByte();
			break;
		case SIG_REGS:
			for(i=0;i<sizeof(regs);i++)	regs[i]=ReadByte();
			break;
		default:
			if(i>=SIG_MEMBLK_BASE && i<(SIG_MEMBLK_BASE+MEMBLK_CT))
			{
				destData=blocks[i-SIG_MEMBLK_BASE];
				destOff=0;
				destSize=16384;
				ReadBlock();
			}
			else
			{
				for(i=0;i<(int32)f;i++)	ReadByte();
			}
			break;
		}

		// check that the number of bytes consumed is correct
		if(j+8+f != (uint32)srcOff)	return(-1);
	}

	return(srcOff);
}

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/***********************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

#include <conio.h>

void Test(void)
{
	uint32		d;
	CState		*s;
	int32		i;
	byte		*b;

	s=new CState();

	for(i=0;i<MEMBLK_CT;i++)
	{
		b=new byte[16384];
		s->SetBlockPtr(i,b);
	}
	s->Reset();

	d=s->Save(NULL,0);
	printf("%ld\n",d);
	getch();
}

