/************************************************************************
 *
 * state.h
 *
 ************************************************************************
 *
 * abstract representation of internal machine state + game config
 *
 ************************************************************************/

#ifndef STATE_H_INCLUDED
#define STATE_H_INCLUDED

/*
 * includes
 */

#include "std.h"

/*
 * defines
 */

#define MEMBLK_48K_ROM		0
#define MEMBLK_48K_4000		1
#define MEMBLK_48K_8000		2
#define MEMBLK_48K_C000		3
#define MEMBLK_128K_ROM		4
#define MEMBLK_128K_RAM0	5
#define MEMBLK_128K_RAM1	6
#define MEMBLK_128K_RAM2	7
#define MEMBLK_128K_RAM3	8
#define MEMBLK_128K_RAM4	9

#define MEMBLK_CT		10


#define REG_A			0
#define REG_F			1
#define REG_B			2
#define REG_C			3
#define REG_D			4
#define REG_E			5
#define REG_H			6
#define REG_L			7
#define REG_A2			8
#define REG_F2			9
#define REG_B2			10
#define REG_C2			11
#define REG_D2			12
#define REG_E2			13
#define REG_H2			14
#define REG_L2			15
#define REG_SPH			16
#define REG_SPL			17
#define REG_PCH			18
#define REG_PCL			19
#define REG_IXH			20
#define REG_IXL			21
#define REG_IYH			22
#define REG_IYL			23
#define REG_I			24
#define REG_R			25
#define REG_IM			26
#define REG_IFF			27
#define REG_BORDER		28
#define REG_MEMMAP		29

#define REG_CT			30


#define IFF_1			1
#define IFF_2			2

#define OPT_VIDEOMODE		0
#define OPT_FRAMESKIP		1
#define OPT_COL_BW		2
#define OPT_BRIGHTNESS		3

#define OPT_CT			4


#define CM_LINEAR			0x00
#define CM_TILEDSCALED			0x01
#define CM_TILEDSCALED_INTERLACED	0x02
#define CM_TILEDSCALED256		0x03
#define CM_TILEDSCALED256_INTERLACED	0x04
#define CM_TILED			0x05
#define CM_TILED_LOCKED			0x06
#define COL_COLOUR			0x00
#define COL_BW				0x01

#define SIG_HEADER			0x55AA0000
#define SIG_TITLE			0x55AA0001
#define SIG_REGS			0x55AA0002
#define SIG_OPTS			0x55AA0003
#define SIG_KEYS			0x55AA0004
#define SIG_MEMBLK_BASE			0x55AA0100
#define SIG_FOOTER			0x55AAFFFF


/*
 * structures/classes
 */

class KeyMap
{
public:
	byte	type;
	byte	port;
	byte	val;
	byte	pad;
};

class CState
{
public:
	void		Reset(void);
	void		WriteMem(int32 blockId,int32 offset,int32 val);
	void		SetReg(int32 regId,int32 val);
	void		SetOpt(int32 optId,int32 val);
	void		SetButton(int32 butId,KeyMap k);
	void		SetTitle(char *t);
	int32		Save(byte *data,int32 maxLen);
	int32		Load(byte *data,int32 maxLen);
	void		SetBlockPtr(int32 blockId,byte *data);
	void		AllocBlockPtrs(void);

	CState(void);

private:
	byte		*blocks[MEMBLK_CT];
	char		title[256];
	KeyMap		keys[16];
	byte		regs[REG_CT];
	byte		opts[OPT_CT];

	byte		*destData;
	int32		destSize;
	int32		destOff;

	byte		*srcData;
	int32		srcSize;
	int32		srcOff;
	bool		srcOverflow;

	void		EmitByte(byte c);
	void		EmitWord(uint16 c);
	void		EmitDword(uint32 c);
	void		EmitBlock(byte *c,int32 len);

	byte		ReadByte(void);
	uint16		ReadWord(void);
	uint32		ReadDword(void);
	void		ReadBlock(void);
};

/*
 * globals
 */

extern CState		st;

/*
 * prototypes
 */

#endif
