/************************************************************************
 *
 * emu.h
 *
 ************************************************************************
 *
 * main emulation routines
 *
 ************************************************************************/

#ifndef EMU_H_INCLUDED
#define EMU_H_INCLUDED

/*
 * includes
 */

#include "std.h"
#include "z80.h"

/*
 * defines
 */

#define KEY_JOY			0x01
#define KEY_KBD			0x02

#define AUDIO_CYCLESPERSAMPLE	218
#define AUDIO_SAMPLESPERFRAME	320
#define AUDIO_CYCLESPERFRAME	(AUDIO_CYCLESPERSAMPLE*AUDIO_SAMPLESPERFRAME)		// should be ~70000
#define AUDIO_FRAMESPERSEC	50
#define AUDIO_FREQ		(AUDIO_FRAMESPERSEC*AUDIO_SAMPLESPERFRAME)

#define HFLAGS_RINC_SHIFT	0x00
#define HFLAGS_RINC_MASK	0x7f
#define HFLAGS_SCR_SHIFT	0x07
#define HFLAGS_SCR_MASK		0x03

/*
 * structures/classes
 */

class KeyMap
{
public:
	byte	type;
	byte	port;
	byte	val;
	byte	pad;
};

class Header
{
public:
	byte	sig[8];
	byte	name[8];
	int32	len;
	uint32	flags;
	KeyMap	keys[16];
};

/*
 * globals
 */

extern byte	*z80Mem;
extern KeyMap	keys[16];
extern byte	*bkpMem;

extern CZ80Asm	*z80Ptr;

/*
 * prototypes
 */

extern void	EmuInit(void);
extern void	EmuCopyScreen(void);
extern void	EmuRun(void);
extern void	EmuTrace(void);
extern void	EmuUpdatePorts(void);
extern int32	EmuLoadZ80(Header *img,bool fetchHeaderInfo);

extern void	EmuBackup(void);
extern void	EmuRestore(void);

extern int32	EmuFindImage(int32 start);
extern int32	EmuGotSig(int32 ad);

#endif
