/***********************************************************************
 *
 * int.cpp
 *
 ***********************************************************************
 *
 * interrupt handling code
 *
 ***********************************************************************/

/*
 * includes
 */

#include "std.h"
#include "int.h"
#include "gba.h"
#include "intasm.h"
#include "mem.h"
#include "emu.h"

/*
 * defines
 */

/*
 * typedefs
 */

/*
 * structs
 */

/*
 * globals
 */

uint32		*intCode;
CIntArgs	*intArgs;
byte		*audioBuffers[2];
int32		audioSize;
int32		empties;

/*
 * function prototypes
 */

/*
 * code
 */


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void IntOpen(uint32 *ic,int32 size)
{
	int32		i,j,k;

	// copy code
	i=(int32)(&IntHandlerCode);
	j=(int32)(&IntHandlerEnd);
	j=((j-i)+3) & (~3);
	ASSERT(size>=j);
	intCode=ic;
	memcpy(intCode,(void*)(&IntHandlerCode),j);

	// compute arg ptr
	k=(int32)(&IntHandlerArgs) - (int32)(&IntHandlerCode);
	intArgs=(CIntArgs*)&intCode[ k>>2 ];

	// set vector
	k=(int32)(&intCode[0]);
	*((int32*)0x03007ffc)=k;

	*((uint16*)(REG_BASE+0x004))=0x08;	// LCD hardware v-blank IE

	// initialise args
	intArgs->vbCount=0;
	intArgs->timerCt=0;

	// set interupt control regs
	*((uint16*)(REG_BASE+0x200))=0x0007;	// vblank+vcount-match+hblank
	*((uint16*)(REG_BASE+0x202))=0xFFFF;	// reset flags
	*((uint16*)(REG_BASE+0x208))=1;		// master IE
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void IntStartAudio(int32 bufSize)
{
	int32		i;

	// allocate 4x as much!!!
	audioSize=bufSize;
	audioBuffers[0]=(byte *)MemAlloc(HEAP_EXTERNAL,bufSize*4);
	audioBuffers[1]=(byte *)MemAlloc(HEAP_EXTERNAL,bufSize*4);
	ASSERT(audioBuffers[0] && audioBuffers[1]);

	for(i=0;i<bufSize;i++)	audioBuffers[0][i]=0;
	for(i=0;i<bufSize;i++)	audioBuffers[1][i]=0;

	intArgs->dmaChannel=0;
	intArgs->dmaAddr0=(uint32)audioBuffers[0];
	intArgs->dmaAddr1=(uint32)audioBuffers[1];

	// set timer0 to ~16KHz	(1049ish cpu ticks)
	*((uint16*)(REG_BASE+0x100))=65536-1049;
	*((uint16*)(REG_BASE+0x102))=0x80;

	// set timer1 to chain off timer 0, divided by bufSize
	*((uint16*)(REG_BASE+0x104))=65536-bufSize;
	*((uint16*)(REG_BASE+0x106))=0xC4;	// interrupt enable, chain off timer0

	// set up DMA1
	*((uint32*)(REG_BASE+0xBC))=(uint32)audioBuffers[0];	// source
	*((uint32*)(REG_BASE+0xC0))=REG_BASE+0xA0;		// dest
	*((uint32*)(REG_BASE+0xC4))=0xB6400000;			// go!

	// init sound
	*((uint16*)(REG_BASE+0x80))=0x0077;	// volume
	*((uint16*)(REG_BASE+0x84))=0x0080;	// master sound enable
	*((uint16*)(REG_BASE+0x82))=0xCB0E;	// reset/clear direct sound A/B

	// enable timer interrupts
	*((uint16*)(REG_BASE+0x200))=0x0017;	// vblank + timer1 + ymatch


}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void IntSubmitBuffer(int32 which)
{
	while(intArgs->dmaChannel!=which);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

