@**********************************************************************************
@
@ intasm.s
@
@**********************************************************************************
@
@ interrupt handler
@
@**********************************************************************************

			.TEXT

			.GLOBAL		IntHandlerCode,IntHandlerArgs,IntHandlerEnd

			.ALIGN
			.ARM


IntHandlerCode:
			@ fetch int status
			ldr		r0,REG_INT_STATUS
			ldr		r1,[r0]


			@ hblank interrupt?
			tsts		r1,#0x0002
			beq		no_horiz_interrupt

			@ which line are we on?
			ldr		r2,yMatchId
			add		r2,r2,#1
			cmps		r2,#0x03
			movge		r2,#0x00
			str		r2,yMatchId

			@ backup r4, set to -> REG_STAT
			str		r4,r4Backup
			ldr		r4,REG_STAT

			@ fetch register information, update regs
			adr		r3,yMatchLine0
			mov		r2,r2,lsl#3
			add		r3,r3,r2
			ldrh		r2,[r3]		@ get new line (REG_STAT) value
			strh		r2,[r4]		@ store it - also disables hlanks!
			ldrh		r2,[r3,#4]	@ get new BG2 bank reg value
			strh		r2,[r4,#8]

			@ restore r4
			ldr		r4,r4Backup
no_horiz_interrupt:


			@ vblank interrupt?
			tsts		r1,#0x0001			@ LCD V-blank interrupt?
			beq		no_vblank_interrupt
			adr		r2,vblankCtr
			ldr		r3,[r2]
			add		r3,r3,#0x1
			str		r3,[r2]
			mov		r2,#0x0001
			strh		r2,[r0]
no_vblank_interrupt:


			@ timer1 (sample interrupt?)
			tsts		r1,#0x0010
			beq		no_timer1_interrupt

			ldr		r2,timerCt
			add		r2,r2,#1
			str		r2,timerCt

			ldr		r2,REG_DMA1_SRC
			ldr		r3,DMA1_STOP
			str		r3,[r2,#8]
			ldr		r3,[r2,#8]
			ldr		r3,[r2,#8]

			adr		r2,dmaChannel
			ldr		r3,[r2]
			eor		r3,r3,#1
			str		r3,[r2],#4
			ldr		r3,[r2,r3,lsl#2]

			ldr		r2,REG_DMA1_SRC
			str		r3,[r2]
			ldr		r3,DMA1_GO
			str		r3,[r2,#8]
no_timer1_interrupt:


			@ y line match interrupt?
			tsts		r1,#0x0004
			beq		no_ymatch_interrupt

			@ enable hblanks!
			ldr		r2,REG_STAT
			ldrh		r3,[r2]
			orr		r3,r3,#16
			strh		r3,[r2]
no_ymatch_interrupt:

			@ reset all interrupts
			strh		r1,[r0]


@ENAB   IRQ    START::START
@-      WIDTH  REPT   SRCINC:
@SRCINC DSTINC:DSTINC -
@-      -      -      -

			bx		lr
REG_INT_STATUS:		.word		0x04000202
REG_DMA1_SRC:		.word		0x040000BC
DMA1_GO:		.word		0xB6400000
DMA1_STOP:		.word		0x76400000
REG_STAT:		.word		0x04000004
r4Backup:		.word		0x00000000

IntHandlerArgs:
vblankCtr:		.word		0x00
dmaChannel:		.word		0x00
dmaAddr0:		.word		0x00
dmaAddr1:		.word		0x00
dmaChain:		.word		0x00
dmaEmpty:		.word		0x00
timerCt:		.word		0x00
dmaAttn:		.word		0x00
yMatchId:		.word		0x00
yMatchLine0:		.word		0x00
yMatchReg0:		.word		0x00
yMatchLine1:		.word		0x00
yMatchReg1:		.word		0x00
yMatchLine2:		.word		0x00
yMatchReg2:		.word		0x00


IntHandlerEnd:

			.END

