/***********************************************************************
 *
 * main.cpp
 *
 ***********************************************************************
 *
 * main program entry, and misc. support functions
 *
 ***********************************************************************/

/*
 * includes
 */

#include "std.h"
#include "font.h"
#include "mem.h"
#include "int.h"
#include "gba.h"
#include "emu.h"
#include "int.h"
#include "copyup.h"
#include "shell.h"
#include "sram.h"

/*
 * defines
 */

/*
 * typedefs
 */

/*
 * structs
 */

/*
 * globals
 */

uint32		*intCodeMem;
byte		*audioBuffer1;
byte		*audioBuffer2;

char		tmp[16];

bool		sramPresent;
bool		sramOk;

/*
 * function prototypes
 */

void		VideoInit(void);
void		Cls(byte col);
void		TinyMemInit(void);
void		SoundInit(void);

/*
 * code
 */



/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void AgbMain(void)
{
	int32		i;
	int32		ad;
	Header		*h;

	// initialise heap
	TinyMemInit();

	// initialise video
	VideoInit();
	Cls(0x00);

	// title message
	DoTitleScreen();

	// scan SRAM
	SRAMScan();

	// initialise shell
	InitShell();

	// initialise interrupt handler
	intCodeMem=(uint32*)MemAlloc(HEAP_EXTERNAL,1024);
	ASSERT(intCodeMem);
	IntOpen(intCodeMem,1024);

	// start audio
	IntStartAudio(AUDIO_SAMPLESPERFRAME);

	// find first image!
	extern int32 _text_end__;
	ad=EmuFindImage(((int32)&_text_end__)-8);
	if(!ad)
	{
		FatalError("NO ROM IMAGE FOUND",__FILE__,__LINE__);
	}

	// init emulator
	EmuInit();

	// init copyup
	CopyupInit();

	// display free IRAM bytes
	i=MemGetFreeBytes(HEAP_INTERNAL);	IntToStr(i,tmp);	DrawString(tmp,100,100,0,255);
	i=MemGetFreeBytes(HEAP_EXTERNAL);	IntToStr(i,tmp);	DrawString(tmp,100,108,0,255);
	IntToStr(sramSize,tmp);			DrawString(tmp,100,118,0,255);

	// load first game
	h=(Header*)ad;
	i=EmuLoadZ80(h,true);
	if(i)	FatalError("BAD ROM IMAGE",__FILE__,__LINE__);
	EmuBackup();

	// main loop
	while(1)
	{
		DoShell();
		EmuRun();
//		EmuTrace();
	}

	ASSERT(0);

	while(1);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void VideoInit(void)
{
	int32		i,j,index;

	// set video mode
	*(uint16 *)REG_LCD = LCD_MODE4 | LCD_BG2_EN;

	// init palette
	for(i=0;i<256;i++)
	{
		int32		r,g,b;

		if(i&128)	index=(i>>0)&7;
		else		index=(i>>3)&7;

		r=0;	g=0;	b=0;
		if(index&1)	b+=0xc0;
		if(index&2)	r+=0xc0;
		if(index&4)	g+=0xc0;
		if(i&64)	{ b*=2; g*=2; r*=2; }
		if(r>0xFF) r=0xFF;
		if(g>0xFF) g=0xFF;
		if(b>0xFF) b=0xFF;

		r>>=3;	g>>=3;	b>>=3;
		j=(b<<10)|(g<<5)|(r<<0);
		((uint16*)PALETTE)[i]=j;
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void Cls(byte col)
{
	uint16		pp;
	int32		i;

	pp=(uint16)col;
	pp|=(pp<<8);

	for(i=0;i<240*160/2;i++)
	{
		((uint16*)VRAM)[i]=pp;
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void FatalError(char *err,char *file,int32 line)
{
	char		tmp[64];
	int32		i;

	VideoInit();
	Cls(0x00);

	for(i=0;file[i];i++) tmp[i]=file[i];
	tmp[i++]=':';
	IntToStr(line,&tmp[i]);

	DrawString(err,8,8,false,0xFF);
	DrawString(tmp,8,16,false,0xFF);
	while(1);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void TinyMemInit(void)
{
	uint32		i;
	extern uint32	_bss_end__;
	i=((uint32)&_bss_end__) + 16;
	MemInit(HEAP_EXTERNAL,(byte *)i,0x02040000-i);
	MemInit(HEAP_INTERNAL,(byte *)0x03000000,32768-512);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void mymemcpy(void *a,void *b,int32 ct)
{
	while(ct--)
	{
		*((byte*)a) = *((byte*)b);
		a=(void*)((uint32)a+1);
		b=(void*)((uint32)b+1);
	}
}


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void mymemset(void *a,byte b,int32 ct)
{
	while(ct--)
	{
		*((byte*)a) = b;
		a=(void*)((uint32)a+1);
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void DoSomething(int q)
{
//	extern void DrawHex(int32 v,int32 x,int32 y,int32 pg,int32 col);
//	Cls(0x00);
//	DrawHex(q,50,50,0,14);
//	while(1);
}
