/***********************************************************************
 *
 * z80.h
 *
 ***********************************************************************
 *
 * structure definitions for asm Z80 emulation core
 *
 ***********************************************************************/

#ifndef Z80_H_INCLUDED
#define Z80_H_INCLUDED

/*
 * includes
 */

#include "std.h"

/*
 * defines
 */

#define IFF_1		0x01
#define IFF_2		0x02
#define HALT		0x80

#define FLAG_C		0x01
#define FLAG_N		0x02
#define FLAG_P		0x04
#define FLAG_H		0x10
#define FLAG_Z		0x40
#define FLAG_S		0x80

/*
 * typedefs
 */

/*
 * structs
 */

class CZ80State
{
public:
	byte		F,A,C,B,E,D,L,H;
	byte		F_,A_,C_,B_,E_,D_,L_,H_;
	uint16		SP,IX,IY,PC;
	byte		IM,IFF;
	byte		I,R;
	byte		lastEI;
	byte		badInst;
	byte		border;
	byte		pad;
	byte		*mem;
	int32		clocks;

	byte		*audioPtr;
	uint32		cyclesPerSample;
	uint32		samplesPerFrame;
	uint32		lastFeValue;
	uint32		rInc;

	// size is 60!
};

class CZ80Asm
{
public:
	CZ80State	state;
	CZ80State	*statePtr;
	byte		*tables;
	uint32		*mainTable;
	uint32		*fastMainTable;
	uint32		retAddr;
	uint32		regBkp[8];

	byte		*feInputPtr;
	byte		*xxInputPtr;
	byte		*keyPtr;
	byte		*keyPortPtr;
	byte		*flashPtr;
	int32		flashCt;
};

/*
 * globals
 */

/*
 * function prototypes
 */

#endif
