@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@ z80.s
@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@ support functions for Z80 emulation
@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

@class CZ80Core
@{
@public:
@	byte		F,A,C,B,E,D,L,H;
@	byte		F_,A_,C_,B_,E_,D_,L_,H_;
@	uint16		SP,IX,IY;
@	uint16		PC;
@	byte		IM,IFF;
@	byte		I,R;
@	byte		lastEI;
@	byte		pad[3];
@	byte		*mem;
@	int32		clocksLeft;
@};

.equ		OFFSET_F,		0
.equ		OFFSET_A,		1
.equ		OFFSET_C,		2
.equ		OFFSET_B,		3
.equ		OFFSET_E,		4
.equ		OFFSET_D,		5
.equ		OFFSET_L,		6
.equ		OFFSET_H,		7
.equ		OFFSET_F_,		8
.equ		OFFSET_A_,		9
.equ		OFFSET_C_,		10
.equ		OFFSET_B_,		11
.equ		OFFSET_E_,		12
.equ		OFFSET_D_,		13
.equ		OFFSET_L_,		14
.equ		OFFSET_H_,		15
.equ		OFFSET_SPL,		16
.equ		OFFSET_SPH,		17
.equ		OFFSET_IXL,		18
.equ		OFFSET_IXH,		19
.equ		OFFSET_IYL,		20
.equ		OFFSET_IYH,		21
.equ		OFFSET_PC,		22
.equ		OFFSET_IM,		24
.equ		OFFSET_IFF,		25
.equ		OFFSET_I,		26
.equ		OFFSET_R,		27
.equ		OFFSET_LASTEI,		28
.equ		OFFSET_BADINST,		29
.equ		OFFSET_MEM,		32
.equ		OFFSET_CLOCKS,		36
.equ		OFFSET_AUDIOPTR,	40
.equ		OFFSET_CYCLESPERSAMPLE,	44
.equ		OFFSET_SAMPLESPERFRAME,	48
.equ		OFFSET_LASTFEVALUE,	52
.equ		OFFSET_RINC,		56

@ size is 56
.equ		Z80_STATE_SIZE,		60


.equ		OFFSET_AF,		OFFSET_F
.equ		OFFSET_BC,		OFFSET_C
.equ		OFFSET_DE,		OFFSET_E
.equ		OFFSET_HL,		OFFSET_L
.equ		OFFSET_AF_,		OFFSET_F_
.equ		OFFSET_BC_,		OFFSET_C_
.equ		OFFSET_DE_,		OFFSET_E_
.equ		OFFSET_HL_,		OFFSET_L_
.equ		OFFSET_SP,		OFFSET_SPL
.equ		OFFSET_IX,		OFFSET_IXL
.equ		OFFSET_IY,		OFFSET_IYL

.equ		TABLE_OFFSET_ROTATE,	0x00000
.equ		TABLE_OFFSET_OR,	0x00100
.equ		TABLE_OFFSET_INC,	0x00200
.equ		TABLE_OFFSET_DEC,	0x00300
.equ		TABLE_OFFSET_DAA,	0x00400
.equ		TABLE_OFFSET_ADD,	0x10000
.equ		TABLE_OFFSET_ADC,	0x20000
.equ		TABLE_OFFSET_SUB,	0x30000
.equ		TABLE_OFFSET_SBC,	0x40000

.equ		IFF_1,			0x01
.equ		IFF_2,			0x02
.equ		IFF_HALTED,		0x80

.equ		FLAG_C,			0x01
.equ		FLAG_N,			0x02
.equ		FLAG_P,			0x04
.equ		FLAG_H,			0x10
.equ		FLAG_Z,			0x40
.equ		FLAG_S,			0x80

		.text
		.code 32
		.align
		.global		Z80CORESTART

@// r14 = fast jump table
@// r13 = core
@// r12 = PC
@// r11 = TABLES
@// r10 = ram
@// r9  = clocks_left
@// r8  = mainTable
@// r7  = 0xFFFF
@// r6  = flags

@// r0  = return address

Z80CORESTART:

z80State:	.space		Z80_STATE_SIZE	
z80StatePtr:	.space		4
flagTables:	.space		4
mainTable:	.space		4
fastMainTable:	.space		4
retAddr:	.space		4
regBkp:		.space		32

feInputPtr:	.space		4
xxInputPtr:	.space		4
keyPtr:		.space		4
keyPortPtr:	.space		4
flashPtr:	.space		4
flashCt:	.space		4

.global Z80CoreEntry
Z80CoreEntry:	str	r1,retAddr
		adr	r1,regBkp
		stmia	r1,{r8-r14}
		
		ldr	r13,z80StatePtr
		ldrh	r12,[r13,#OFFSET_PC]
		ldr	r10,[r13,#OFFSET_MEM]
		ldrb	r6,[r13,#OFFSET_F]
		ldrb    r5,[r13,#OFFSET_A]
		mov	r7,#0x10000
		sub	r7,r7,#1

		ldr	r8,mainTable
		ldr	r14,fastMainTable
		ldr	r11,flagTables
		adr	r0,main_loop

		@ first time, r9 is read from CLOCKS!
		@ on subsequent occasions CYCLESPERSAMPLE is added...
		ldr	r9,[r13,#OFFSET_CLOCKS]

main_cont:	subs	r9,r9,#0
main_loop:	@ LAST INSTRUCTION TO SET FLAGS MUST BE CLOCK-DECREMENT!
		blt	break_out
		and	r12,r12,r7
		ldrb	r4,[r10,r12]
		add	r12,r12,#1
		ldr	r15,[r14,r4,lsl#2]

break_out:	@ store audio sample, increment ptr
		ldr	r2,[r13,#OFFSET_AUDIOPTR]
		ldr	r1,[r13,#OFFSET_LASTFEVALUE]
		strb	r1,[r2],#1
		str	r2,[r13,#OFFSET_AUDIOPTR]

		@ dec ctr
		ldr	r1,[r13,#OFFSET_SAMPLESPERFRAME]
		subs	r1,r1,#1
		beq	break_out_really
		str	r1,[r13,#OFFSET_SAMPLESPERFRAME]

		ldr	r3,[r13,#OFFSET_CYCLESPERSAMPLE]
		add	r9,r9,r3
		b	main_cont

break_out_really:
		strh	r12,[r13,#OFFSET_PC]
		str	r9,[r13,#OFFSET_CLOCKS]
		strb	r6,[r13,#OFFSET_F]
		strb    r5,[r13,#OFFSET_A]

		@ restore regs
return:		adr	r1,regBkp
		ldmia	r1,{r8-r14}
		ldr	r1,retAddr
		bx	r1

.global DoCB
DoCB:		ldrb	r4,[r10,r12]
		add	r12,r12,#1
		add	r1,r4,#0x100		@ CB=table 1
		ldr	r15,[r8,r1,lsl#2]
.global DoED
DoED:		ldrb	r4,[r10,r12]
		add	r12,r12,#1
		add	r1,r4,#0x200		@ ED=table 2
		ldr	r15,[r8,r1,lsl#2]
.global DoDD
DoDD:		ldrb	r4,[r10,r12]
		add	r12,r12,#1
		add	r1,r4,#0x300		@ DD=table 3
		ldr	r15,[r8,r1,lsl#2]
.global DoFD
DoFD:		ldrb	r4,[r10,r12]
		add	r12,r12,#1
		add	r1,r4,#0x400		@ FD=table 4
		ldr	r15,[r8,r1,lsl#2]
.global DoDDCB
DoDDCB:		add	r12,r12,#1
		ldrb	r4,[r10,r12]
		sub	r12,r12,#1
		add	r0,r4,#0x500		@ DDCB=table 5
		ldr	r1,[r8,r0,lsl#2]
		adr	r0,DoDDCBReturn
		mov	r15,r1
DoDDCBReturn:	add	r12,r12,#1
		adr	r0,main_loop
		b	main_loop

.global DoFDCB
DoFDCB:		add	r12,r12,#1
		ldrb	r4,[r10,r12]
		sub	r12,r12,#1
		add	r0,r4,#0x600		@ FDCB=table 6
		ldr	r1,[r8,r0,lsl#2]
		adr	r0,DoDDCBReturn
		mov	r15,r1

.global DoBAD
DoBAD:		mov	r1,#1
		strb	r1,[r13,#OFFSET_BADINST]
		b	break_out_really


		@
		@ process IN port
		@ r1=HI r2=LO r3=DEST
		@
DoInput:	cmps	r2,#0xfe
		ldreq	r4,feInputPtr
		ldreqb	r4,[r4,r1]
		ldrne	r4,xxInputPtr
		ldrneb	r4,[r4,r2]
		strb	r4,[r3]
		ldrb	r3,[r11,r4]	@ fetch from rotate table
		and	r3,r3,#~FLAG_C
		and	r6,r6,#FLAG_C
		orr	r6,r6,r3
		mov	r1,r14			@ copy return address
		ldr	r14,fastMainTable	@ restore r14- it was trashed getting here
		mov	r15,r1			@ return

		@
		@ process OUT val,port
		@ r1=val, r2=port
		@
DoOutput:	cmps	r2,#0xfe		@ speaker?
		bne	done_output

		@ convert value
		tsts	r1,#16
		moveq	r1,#128
		movne	r1,#0

		@ store last output value
		strb	r1,[r13,#OFFSET_LASTFEVALUE]

done_output:	mov	r1,r14			@ copy return address
		ldr	r14,fastMainTable	@ restore r14- it was trashed getting here
		mov	r15,r1			@ return


		@ convert keyboard status into 256 0xFE port input values
		@ (you can read several groups of keys at once!)

		.global BuildKeyMasks
BuildKeyMasks:
		str	r1,retAddr
		adr	r1,regBkp
		stmia	r1,{r8-r14}

		ldr	r0,keyPtr
		ldr	r1,keyPortPtr

		@ read 8 row bitfields
		ldrb	r2,[r0,#0]
		ldrb	r3,[r0,#1]
		ldrb	r4,[r0,#2]
		ldrb	r5,[r0,#3]
		ldrb	r6,[r0,#4]
		ldrb	r7,[r0,#5]
		ldrb	r8,[r0,#6]
		ldrb	r9,[r0,#7]

		mov	r10,#0x00

plp:		mov	r11,#0xFF
		tsts	r10,#0x01
		andeq	r11,r11,r2
		tsts	r10,#0x02
		andeq	r11,r11,r3
		tsts	r10,#0x04
		andeq	r11,r11,r4
		tsts	r10,#0x08
		andeq	r11,r11,r5
		tsts	r10,#0x10
		andeq	r11,r11,r6
		tsts	r10,#0x20
		andeq	r11,r11,r7
		tsts	r10,#0x40
		andeq	r11,r11,r8
		tsts	r10,#0x80
		andeq	r11,r11,r9
		strb	r11,[r1,r10]
		add	r10,r10,#1
		cmps	r10,#0x100
		blt	plp

		b	return

		@
		@ mess with FLASH attributes
		@
		.global		DoFlash
DoFlash:	str	r1,retAddr
		adr	r1,regBkp
		stmia	r1,{r8-r14}
		
		ldr	r0,flashPtr
		ldr	r1,flashCt
att_loop:	ldrb	r2,[r0],#1
		tsts	r2,#0x80
		bne	att_flash
		subs	r1,r1,#1
		bne	att_loop
		b	return
att_flash:	mov	r3,r2,lsr#3
		and	r3,r3,#7
		mov	r4,r2,lsl#3
		and	r4,r4,#(7<<3)
		and	r2,r2,#0xC0
		orr	r2,r2,r3
		orr	r2,r2,r4
		strb	r2,[r0,#-1]
		subs	r1,r1,#1
		bne	att_loop
		b	return



		@ code follows...

		.include "z80core.s"
		.include "z80tab.s"

