/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/*
 * includes
 */

#include "std.h"
#include "mem.h"

/*
 * defines
 */

#define STACK_BYTES		4096
#define SAFE_BYTES		16

#define HEAP_CT			2

#define SIG_FREE		0xAAAA
#define SIG_ALLOCATED		0x5555
#define SIG_FIRST		0x5A5A
#define SIG_LAST		0xA5A5

/*
 * structures/classes
 */

class CMemBlock
{
public:
	uint16		sig;
	uint16		paras;
	uint16		line;
	byte		heapId;
	byte		pad;
	char		file[8];
};

/*
 * globals
 */

CMemBlock*	heaps[HEAP_CT];

/*
 * prototypes
 */

/*
 * code
 */

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void MemInit(int32 heapId,byte* start,uint32 len)
{
	CMemBlock		*mem;
	uint32			st;

	ASSERT(heapId>=0 && heapId<HEAP_CT);

	st=(uint32)start;
	while(st&0xF)	{ st++; len--; }
	len>>=4;

	mem=(CMemBlock *)st;
	mem->paras=0;
	mem->sig=SIG_FIRST;
	mem->line=0;
	mem->file[0]=0;
	mem->heapId=(byte)heapId;

	heaps[heapId]=mem;

	mem=(CMemBlock *)(st + 16);
	mem->paras=(uint16)(len-3);
	mem->sig=SIG_FREE;
	mem->line=0;
	mem->file[0]=0;
	mem->heapId=(byte)heapId;

	mem=(CMemBlock *)(st + ((len-1)<<4));
	mem->sig=SIG_LAST;
	mem->paras=0;
	mem->line=0;
	mem->file[0]=0;
	mem->heapId=(byte)heapId;
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

byte* MemDoAlloc(int32 heapId,uint32 len,char *file,int16 line)
{
	CMemBlock		*mem;
	CMemBlock		*best;
	int32			i;

	ASSERT(heapId>=0 && heapId<HEAP_CT);

	len=(len+15)>>4;
	mem=heaps[heapId];
	best=NULL;
	while(1)
	{
		if(mem->sig == SIG_LAST)	break;

		if(mem->sig == SIG_FREE)
		{
			if(mem->paras >= len)
			{
				if(!best)	best=mem;
				else
				{
					if(mem->paras < best->paras)
					{
						best=mem;
					}
				}
			}
		}

		mem=(CMemBlock*)((uint32)mem + ((mem->paras+1)<<4));
	}

	if(!best)	return(NULL);

	if(len==(uint32)best->paras-1)	len++;

	if(best->paras > len)
	{
		mem=(CMemBlock*)((uint32)best+((len+1)<<4));
		mem->sig=SIG_FREE;
		mem->paras=(uint16)(best->paras-len-1);
	}
	best->paras=(uint16)len;

	for(i=0;i<8;i++)	best->file[i]=file[i];
	best->line=line;
	best->sig=SIG_ALLOCATED;
	best->heapId=(byte)heapId;

	return(((byte *)best)+16);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void MemDoFree(byte *data,char *file,int16 line)
{
	CMemBlock		*mem,*last;
	int32			heapId;

	mem=(CMemBlock*)(data-16);
//	ASSERT(mem->sig==SIG_ALLOCATED);
	if(mem->sig!=SIG_ALLOCATED)	FatalError("Free bad block",file,line);

	mem->sig=SIG_FREE;
	heapId=mem->heapId;

	// merge free blocks
	last=heaps[heapId];
	for(mem=heaps[heapId];mem->sig!=SIG_LAST;mem=(CMemBlock*)((uint32)mem+((mem->paras+1)<<4)))
	{
		ASSERT(mem->sig==SIG_ALLOCATED || mem->sig==SIG_FREE || mem->sig==SIG_FIRST);
		if(mem->sig==SIG_FREE && last->sig==SIG_FREE)
		{
			last->paras += mem->paras+1;
			mem=last;
		}
		last=mem;
	}
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

void StdMemInit(void)
{
	uint32		i;
	extern uint32	_bss_end__;

	i=((uint32)&_bss_end__) + SAFE_BYTES;

	MemInit(HEAP_EXTERNAL,(byte *)i,0x02040000-i);
	MemInit(HEAP_INTERNAL,(byte *)0x03000000,32768-STACK_BYTES);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

int32 MemGetFreeBytes(int32 heapId)
{
	CMemBlock		*mem;
	int32			total;

	ASSERT(heapId>=0 && heapId<HEAP_CT);

	total=0;
	mem=heaps[heapId];
	while(1)
	{
		if(mem->sig == SIG_LAST)	break;
		if(mem->sig == SIG_FREE)	total+=mem->paras << 4;
		mem=(CMemBlock*)((uint32)mem + ((mem->paras+1)<<4));
	}
	return(total);
}

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

/************************************************************************
 *
 *
 *
 ************************************************************************
 *
 *
 *
 ************************************************************************/

