/***********************************************************************
 *
 * z80.cpp
 *
 ***********************************************************************
 *
 * program to generate ARM32 Z80 emulation core from templates
 *
 ***********************************************************************/

/*
 * includes
 */

#include "std.h"
#include "emit.h"
#include <conio.h>

/*
 * defines
 */

#define EMIT_ASM

/*
 * typedefs
 */

/*
 * structs
 */

class CInstVar
{
public:
	char		code;
	int32		byteId;
	byte		mask;
	int32		bitCt,shift;

	int32		tmpVal;

	bool		read,write;
};

class CInstruction
{
public:
	int32		len;

	char		*opString,*argString;
	char		*iName;
	int32		tStates;

	int32		instVarCt;
	CInstVar	instVars[4];

	int32		byteCt;
	byte		bytes[8];
	byte		masks[8];

	int32		hlIndir;

	void		Dump(void)
	{
		int32		i,j,k;

		printf("[%s] [%s] :\n",opString,argString);

		printf("%d bytes...\n",byteCt);
		for(i=0;i<byteCt;i++)
		{
			printf("  ");
			for(j=0;j<8;j++)
			{
				if(masks[i] & (128>>j))	printf("%d",(bytes[i]&(128>>j))?1:0);
				else			printf("?");
			}
		}
		printf("\n");
		for(i=0;i<byteCt;i++)
		{
			printf("  ");
			for(j=0;j<8;j++)
			{
				for(k=0;k<instVarCt;k++)
				{
					if(instVars[k].byteId!=i)	continue;
					if(instVars[k].mask & (128>>j))	break;
				}
				if(k==instVarCt)	printf(".");
				else			printf("%c",instVars[k].code);
			}
		}
		printf("\n");
	}
};

/*
 * globals
 */


CInstruction*	mainTable[256];
CInstruction*	cbTable[256];
CInstruction*	ddTable[256];
CInstruction*	edTable[256];
CInstruction*	fdTable[256];

#define MAX_INSTR	1024
CInstruction	insts[MAX_INSTR];
int32		instCt;

byte		dupe[256];

char*		reg8Names[]={	"B","C","D","E","H","L","(HL)","A" };
char*		reg16Names[]={	"BC","DE","HL","SP" };
char*		stack16Names[]={ "BC","DE","HL","AF" };

char*		reg8NamesIX[]={	"B","C","D","E","IXh","IXl","(IX+?)","A" };
char*		reg16NamesIX[]={	"BC","DE","IX","SP" };
char*		stack16NamesIX[]={ "BC","DE","IX","AF" };
char*		reg8NamesIY[]={	"B","C","D","E","IYh","IYl","(IY+?)","A" };
char*		reg16NamesIY[]={	"BC","DE","IY","SP" };
char*		stack16NamesIY[]={ "BC","DE","IY","AF" };

bool		ixMode,iyMode;
int32		disp;
bool		usedDisp;

/*
 * function prototypes
 */

void		Insert(CInstruction **tab,int32 off,CInstruction *i);
int32		ReadInstruction(char *line,CInstruction *s);
int32		Disassemble(char *dest,byte *data);

void		EmitFunction(byte pref,byte op,byte op2);
void		Crunch(char *ii,char *oo);
char *		GetFn(byte a,byte b,byte c);


/*
 * code
 */


void EmitProto(byte a,byte b,byte c)
{
	char		*s;
	char		temp[128];

	s=GetFn(a,b,c);
	if(!s)
	{
#ifdef EMIT_ASM
		fprintf(wr,"    .word   0\t\t@ %02x%02x%02x\n",a,b,c);
#else
		fprintf(wr,"    NULL,\t\t// %02x%02x%02x\n",a,b,c);
#endif
	}
	else
	{
		char	*c;
#ifdef EMIT_ASM
		strcpy(temp,s);
		c=strstr(temp,":");
		*c=0;
		fprintf(wr,"    .word   %s\t\t@ %02x%02x%02x\n",temp,a,b,c);
#else
		strcpy(temp,&s[5]);
		c=strstr(temp,"(");
		*c=0;
		fprintf(wr,"    %s,\t\t// %02x%02x%02x\n",temp,a,b,c);
#endif
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void main(int argc, char **argv)
{
	CInstruction		q;
	FILE			*stream;
	char			tmp[256];
	int32			i,j,k;

//	void	Test(void);
//	Test();
//	exit(1);

	instCt=0;

	// read instructions
	stream=fopen("inst3.txt","r");
	while(1)
	{
		fgets(tmp,sizeof(tmp),stream);
		if(feof(stream))	break;

		if(strstr(tmp,"#"))	*(strstr(tmp,"#"))=0;

		i=ReadInstruction(tmp,&insts[instCt]);
		if(i)	continue;

		instCt++;
	}
	fclose(stream);

	// clear tables
	memset(mainTable,0,sizeof(mainTable));
	memset(cbTable,0,sizeof(cbTable));
	memset(ddTable,0,sizeof(ddTable));
	memset(edTable,0,sizeof(edTable));
	memset(fdTable,0,sizeof(fdTable));
	memset(dupe,0,sizeof(dupe));

	// populate tables
	for(i=0;i<256;i++)
	{
		for(j=0;j<instCt;j++)
		{
			if(insts[j].bytes[0]==0xCB && insts[j].byteCt>1)
			{
				if(insts[j].bytes[1]==(i&insts[j].masks[1]))
				{
					Insert(cbTable,i,&insts[j]);
				}
			}
			else if(insts[j].bytes[0]==0xDD)
			{
				printf("shit\n"); exit(1);
			}
			else if(insts[j].bytes[0]==0xED && insts[j].byteCt>1)
			{
				if(insts[j].bytes[1]==(i&insts[j].masks[1]))
				{
					Insert(edTable,i,&insts[j]);
				}
			}
			else if(insts[j].bytes[0]==0xFD)
			{
				printf("shit\n"); exit(1);
			}
			else
			{
				if(insts[j].bytes[0]==(i&insts[j].masks[0]))
				{
					Insert(mainTable,i,&insts[j]);
				}
			}
		}
	}

#if 0
	for(i=0;i<256;i++)
	{
		printf("%-5ld[%02x] ",i,i);
		if(mainTable[i])	printf("%-20s",mainTable[i]->iName);
		else			printf("                    ");
		if(cbTable[i])		printf("%-20s",cbTable[i]->iName);
		else			printf("                    ");
		if(edTable[i])		printf("%-20s",edTable[i]->iName);
		else			printf("                    ");
		
		printf("\n");
	}
#endif


	wr=fopen("temp.cpp","w");
	for(i=0;i<256;i++)		EmitFunction(0x00,i,0x00);
	for(i=0;i<256;i++)		EmitFunction(0xCB,i,0x00);
	for(i=0;i<256;i++)		EmitFunction(0xED,i,0x00);
	for(i=0;i<256;i++)		EmitFunction(0xDD,i,0x00);
	for(i=0;i<256;i++)		EmitFunction(0xFD,i,0x00);
	for(i=0;i<256;i++)		EmitFunction(0xDD,0xCB,i);
	for(i=0;i<256;i++)		EmitFunction(0xFD,0xCB,i);
	fclose(wr);

	Crunch("temp.cpp","z80core.s");
	remove("temp.cpp");
	wr=fopen("z80tab.s","w");

#ifndef EMIT_ASM
	fprintf(wr,"#include \"pc.cpp\"\n");
	fprintf(wr,"void (*mainTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(i,0x00,0x00);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*cbTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xCB,i,0x00);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*edTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xED,i,0x00);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*ddTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xDD,i,0x00);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*fdTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xFD,i,0x00);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*ddcbTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xDD,0xCB,i);
	fprintf(wr,"};\n");
	fprintf(wr,"void (*fdcbTable[256])(void)={\n");
	for(i=0;i<256;i++)		EmitProto(0xFD,0xCB,i);
	fprintf(wr,"};\n");
#else
	fprintf(wr,".global JT_MAIN\nJT_MAIN:\n");
	for(i=0;i<256;i++)		EmitProto(i,0x00,0x00);
	fprintf(wr,".global JT_CB\nJT_CB:\n");
	for(i=0;i<256;i++)		EmitProto(0xCB,i,0x00);
	fprintf(wr,".global JT_ED\nJT_ED:\n");
	for(i=0;i<256;i++)		EmitProto(0xED,i,0x00);
	fprintf(wr,".global JT_DD\nJT_DD:\n");
	for(i=0;i<256;i++)		EmitProto(0xDD,i,0x00);
	fprintf(wr,".global JT_FD\nJT_FD:\n");
	for(i=0;i<256;i++)		EmitProto(0xFD,i,0x00);
	fprintf(wr,".global JT_DDCB\nJT_DDCB:\n");
	for(i=0;i<256;i++)		EmitProto(0xDD,0xCB,i);
	fprintf(wr,".global JT_FDCB\nJT_FDCB:\n");
	for(i=0;i<256;i++)		EmitProto(0xFD,0xCB,i);
#endif

	fclose(wr);
	exit(1);



	byte	*data;
	data=new byte[16384];
	stream=fopen("48.rom","rb");
	if(!stream) exit(1);
	fread(data,1,16384,stream);
	fclose(stream);
	i=0;

	while(1)
	{
		j=Disassemble(tmp,&data[i]);
		printf("%04x : %s\n",i,tmp);
		i+=j;
		getch();
	}
	printf("%s\n",tmp);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 ReadInstruction(char *line,CInstruction *s)
{
	int32		i,j,k;
	char		tmp[128];
	char		*c,*d;
	char		readList[16];
	char		writeList[16];

	int32		readListLen,writeListLen;

	for(i=0;line[i];i++) if(line[i]=='\t') line[i]=' ';

	readListLen=0;
	writeListLen=0;
	
	i=strlen(line);
	if(line[i-1]=='\n')	line[i-1]=0;

	// read bytes
	memset(s,0,sizeof(CInstruction));
	c=line;
	while(1)
	{
		for(i=0;i<8;i++)
		{
			if(c[i]==' ' || c[i]==0)	break;
		}
		if(i!=8)	break;
		if(c[i]!=' ')	break;

		memcpy(tmp,c,8);
		tmp[8]=0;

		c+=9;

		s->bytes[s->byteCt]=0;
		s->masks[s->byteCt]=0;

		for(i=0;i<8;i++)
		{
			if(tmp[i]=='1')
			{
				s->bytes[s->byteCt] |= (128>>i);
				s->masks[s->byteCt] |= (128>>i);
			}
			else if(tmp[i]=='0')
			{
				s->masks[s->byteCt] |= (128>>i);
			}
			else
			{
				for(j=0;j<s->instVarCt;j++)
				{
					if(s->instVars[j].code == tmp[i])	break;
				}
				if(j==s->instVarCt)
				{
					s->instVars[s->instVarCt].code=tmp[i];
					s->instVars[s->instVarCt].byteId=s->byteCt;
					s->instVarCt++;
				}
				s->instVars[j].mask |= (128>>i);
			}
		}
		s->byteCt++;
	}
	if(!s->byteCt)	return(-1);

	i=sscanf(c,"%ld",&s->tStates);
	if(i!=1)	return(-1);

	i=0;
	j=0;
	for(d=&c[strlen(c)];d!=c;d--)
	{
		if(*d==' ' || (*d)==0 || *d=='\t')
		{
			*d=0;
			if(i)
			{
				d++;
				if(d[0]=='%')
				{
					if(d[1]=='r')	{ readList[readListLen++]=d[2]; }
					if(d[1]=='w')	{ writeList[writeListLen++]=d[2]; }
					j--;
				}
				else if(d[0]=='+')
				{
					s->hlIndir=atoi(&d[1]);
					j--;
				}
				else if(j==0)
				{
					s->iName=strdup(d);
				}
				else if(j==1)
				{
					s->argString=strdup(d);
				}
				else if(j==2)
				{
					s->opString=strdup(d);
				}

				j++;
				d--;
			}
			i=0;
		}
		if(*d!=' ' && (*d)!=0)	i=1;
	}
	if(s->opString[0] >= '0' && s->opString[0] <= '9')
	{
		s->opString=s->argString;
		s->argString=NULL;
	}

	// fix up read/write flags, compute shifts
	for(i=0;i<s->instVarCt;i++)
	{
		for(j=0;j<readListLen;j++)
		{
			if(readList[j]==s->instVars[i].code)	s->instVars[i].read=true;
			if(writeList[j]==s->instVars[i].code)	s->instVars[i].write=true;
		}

		k=s->instVars[i].mask;
		for(j=0;!(k&1);j++)	k>>=1;
		s->instVars[i].shift=j;
	}



	return(0);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void Insert(CInstruction **tab,int32 off,CInstruction *i)
{
	if(tab[off])
	{
		printf("dupe: %s %s / %s %s\n",
			i->opString,i->argString,
			tab[off]->opString,tab[off]->argString);
	}
	tab[off]=i;
}


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

void Dump(char *dest,CInstruction *w,byte *data)
{
	int32		loVal,hiVal;
	int32		i,j,k,ww;
	char		tmp[128];

	char		**r8Names;
	char		**r16Names;
	char		**s16Names;


	sprintf(dest,"%-8s ",w->opString);
	loVal=-1;	hiVal=-1;

	for(i=0;i<w->instVarCt;i++)
	{
		ww=w->instVars[i].byteId;
		if((ixMode || iyMode) && ww)	ww++;
		w->instVars[i].tmpVal=data[ ww ] & w->instVars[i].mask;
		w->instVars[i].tmpVal>>=w->instVars[i].shift;
	}
	for(i=0;i<w->instVarCt;i++)
	{
		if(w->instVars[i].code=='l')	loVal=w->instVars[i].tmpVal;
		if(w->instVars[i].code=='h')	hiVal=w->instVars[i].tmpVal;
	}
	if(loVal!=-1 && hiVal!=-1)
	{
		hiVal=(hiVal<<8) + loVal;
	}

	r8Names=reg8Names;
	r16Names=reg16Names;
	s16Names=stack16Names;

	if(ixMode)
	{
		r8Names=reg8NamesIX;
		r16Names=reg16NamesIX;
		s16Names=stack16NamesIX;
	}
	if(iyMode)
	{
		r8Names=reg8NamesIY;
		r16Names=reg16NamesIY;
		s16Names=stack16NamesIY;
	}

	tmp[0]=0;
	if(w->argString)
	{
		for(i=0;w->argString[i];i++)
		{
			if(w->argString[i]=='!')
			{
				if(ixMode)	strcat(tmp,"IX");
				else if(iyMode)	strcat(tmp,"IY");
				else		strcat(tmp,"HL");
				continue;
			}
			if(w->argString[i]=='@')
			{
				char		arse[128];

				if(ixMode)	sprintf(arse,"(IX+#%02x)",((signed char)disp)&0xFF);
				else if(iyMode)	sprintf(arse,"(IY+#%02x)",((signed char)disp)&0xFF);
				else		sprintf(arse,"(HL)");
				strcat(tmp,arse);

				if(ixMode || iyMode)	usedDisp=true;
				continue;
			}

			// normal char?
			if(w->argString[i]<'a' || w->argString[i]>'z')
			{
				j=strlen(tmp);
				tmp[j++]=w->argString[i];
				tmp[j]=0;
				continue;
			}

			// subst
			for(j=0;j<w->instVarCt;j++)
			{
				if(w->instVars[j].code!=w->argString[i])	continue;
				switch(w->argString[i])
				{
				case 'x':
					if(w->instVars[j].tmpVal)	strcat(tmp,"C");
					else				strcat(tmp,"NC");
					break;
				case 'y':
					if(w->instVars[j].tmpVal)	strcat(tmp,"Z");
					else				strcat(tmp,"NZ");
					break;
				case 'r':
				case 's':
					strcat(tmp,r8Names[w->instVars[j].tmpVal]);
					break;
				case 'd':
					strcat(tmp,r16Names[w->instVars[j].tmpVal]);
					break;
				case 'q':
					strcat(tmp,s16Names[w->instVars[j].tmpVal]);
					break;
				case 'n':
					char	arse[16];
					sprintf(arse,"#0x%02x",w->instVars[j].tmpVal);
					strcat(tmp,arse);
					break;
				case 'h':
					sprintf(&tmp[strlen(tmp)],"#0x%04x",hiVal);
					i+=2;
					break;

				default:
					strcat(tmp,"***");
					break;
				}
				break;
			}
		}

	}

	strcat(dest,tmp);
}


/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32 Disassemble(char *dest,byte *data)
{
	char		tmp[128];
	int32		i,j;
	CInstruction	*w;
	int32		q;

	ixMode=false;
	iyMode=false;
	usedDisp=false;
	
	disp=-1;
	q=data[0];
	w=mainTable[q];
	if(w)
	{
		Dump(dest,w,data);
		return(w->byteCt);
	}
	else
	{
		if(data[0]==0xCB)
		{
			q=data[1];
			w=cbTable[q];
			if(w)
			{
				Dump(dest,w,&data[0]);
				return(w->byteCt);
			}
		}
		else if(data[0]==0xED)
		{
			q=data[1];
			w=edTable[q];
			if(w)
			{
				Dump(dest,w,&data[0]);
				return(w->byteCt);
			}
		}
		else if(data[0]==0xDD || data[0]==0xFD)
		{
			if(data[0]==0xDD)	ixMode=true;
			else			iyMode=true;

			data++;
			q=data[0];
			if(q==0xCB)
			{
				data--;
				q=data[2];
				w=cbTable[q];
				if(w)
				{
					disp=data[1];
					Dump(dest,w,&data[0]);
					return(w->byteCt+1);
				}
			}
			else
			{
				w=mainTable[q];
				if(w)
				{
					disp=data[1];
					Dump(dest,w,data);
					return(w->byteCt+1+(usedDisp?1:0));
				}
			}
		}
	}


	return(-1);
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32	labCount=0;

void EmitFunction(byte pref,byte op,byte op2)
{
	int32		i,j,k;
	CInstruction	*w;
	int32		vals[3];
	int32		v;
	int32		opOff;
	byte		prefCode;
	byte		tmpData[16];
	char		tmp[128];
	byte		oo;
	char		iName[128];
	int32		regOff;
	bool		superNasty;
	bool		doublePref;

	superNasty=false;

	// get instruction
	opOff=0;

	w=NULL;
	prefCode=PREFIX_NONE;

	doublePref=false;

	switch(pref)
	{
	case 0x00:
		w=mainTable[op];
		break;
	case 0xCB:
		w=cbTable[op];
		opOff++;
		break;
	case 0xED:
		w=edTable[op];
		opOff++;
		break;
	case 0xDD:
		if(op==0xCB)
		{
			opOff++;
			w=cbTable[op2];
			doublePref=true;
		}
		else	w=mainTable[op];
		prefCode=PREFIX_IX;
		break;
	case 0xFD:
		if(op==0xCB)
		{
			opOff++;
			w=cbTable[op2];
			doublePref=true;
		}
		else	w=mainTable[op];
		prefCode=PREFIX_IY;
		break;
	}

	if(!w)
	{
		printf("missing %02x:%02x:%02x\n",pref,op,op2);
		return;
	}

	memset(tmpData,0x55,sizeof(tmpData));
	i=0;
	if(pref)	tmpData[i++]=pref;
	tmpData[i++]=op;
	tmpData[i++]=op2;
	Disassemble(tmp,tmpData);
#ifdef EMIT_ASM
	fprintf(wr,"@// %s [%s]\n",tmp,w->iName);
#else
	fprintf(wr,"// %s [%s]\n",tmp,w->iName);
#endif

//	sprintf(currentLabel,"lab_%04x",labCount++);

	if(!pref)
	{
#ifdef EMIT_ASM
		fprintf(wr,"Emulate_%02x:\n",op);
#else
		fprintf(wr,"void Emulate_%02x(void) {\n",op);
#endif
	}
	else
	{
		if(pref==0xDD || pref==0xFD)
		{
#ifdef EMIT_ASM
			if(op==0xCB)	fprintf(wr,"Emulate_%02x_%02x_%02x:\n",pref,op,op2);
			else		fprintf(wr,"Emulate_%02x_%02x:\n",pref,op);
#else
			if(op==0xCB)	fprintf(wr,"void Emulate_%02x_%02x_%02x(void) {\n",pref,op,op2);
			else		fprintf(wr,"void Emulate_%02x_%02x(void) {\n",pref,op);
#endif
		}
		else
		{
#ifdef EMIT_ASM 
			fprintf(wr,"Emulate_%02x_%02x:\n",pref,op);
#else
			fprintf(wr,"void Emulate_%02x_%02x(void) {\n",pref,op);
#endif		
		}
	}

	oo=op;
	if(pref==0xDD || pref==0xFD)
	{
		if(op==0xCB || op==0xED)	op=op2;
	}
	instNum=op;

	// get emit-time args
	v=0;
	for(i=0;i<w->instVarCt;i++)
	{
		if(w->instVars[i].byteId!=opOff)	continue;
		vals[v++]=(op & w->instVars[i].mask) >> (w->instVars[i].shift);
	}

	// copy family
	strcpy(iName,w->iName);

	// special DD-CB-?? or FD-CB-??
#if 0
	if((pref==0xDD || pref==0xFD) && (oo==0xCB))
	{
		// change instruction group to (IX+d) or (IY+d)
		char	*c;
		c=strstr(iName,"BIT_B_");	// leave BIT's alone
		if(!c)
		{
			c=strstr(iName,"_REG");
			if(c)
			{
				strcpy(c,"_IHL");
				superNasty=true;
			}

			if(strstr(iName,"SET_") || strstr(iName,"RES_"))	regOff=1;
			else							regOff=0;
		}
	}
#endif

	// find emit function
#ifdef EMIT_ASM
	extraClocks=0;
	for(i=0;asmEmitDefs[i].name;i++)
	{
		if(!stricmp(asmEmitDefs[i].name,iName))
		{
			asmEmitDefs[i].emitFunc(prefCode,vals[0],vals[1],vals[2]);

			fprintf(wr,"    subs   r9,r9,#%ld\n",w->tStates+extraClocks);

			// super-nasty DDCB/FDCB
			if(superNasty)
			{
				fprintf(wr,"    strb   r1,[r13,OFFSET_%s]     @ undocumented writeback\n",ireg8StrNormal[vals[regOff]]);
			}
//			fprintf(wr,"    bx r0\n");
			fprintf(wr,"    mov r15,r0\n");
			break;
		}
	}
	if(!asmEmitDefs[i].name)
	{
		fprintf(wr,"    b  DoBAD\n");
		printf("missing %s\n",iName);
	}
#else
	for(i=0;emitDefs[i].name;i++)
	{
		if(!stricmp(emitDefs[i].name,iName))
		{
			emitDefs[i].emitFunc(prefCode,vals[0],vals[1],vals[2]);
			fprintf(wr,"core->clocks += %ld;\n",w->tStates);
			break;
		}
	}
	// super-nasty DDCB/FDCB
	if(superNasty)
	{
		fprintf(wr,"printf(\"undoc\\n\");\n");
		fprintf(wr,"core->%s = (byte)nv;   // undocumented writeback\n",reg8StrNormal[vals[regOff]]);
		fprintf(wr,"DISASM(\"->%s\");\n",reg8StrNormal[vals[regOff]]);
	}
	fprintf(wr,"}\n");
#endif

}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

#define MAX_LINES	20
class CFunc
{
public:
	char*	com;
	char*	name;
	int32	lineCt;
	char*	lines[MAX_LINES];

	CFunc(void)
	{
		com=NULL;
		name=NULL;
		lineCt=0;
	}
};

class CMapping
{
public:
	char		*name;
	char		*dest;
};

#define MAX_FUNCS	4096
int32		funcCt;
CFunc		*funcs[MAX_FUNCS];

int32		numMaps=0;
CMapping	maps[MAX_FUNCS];

void AddMapping(char *a,char *b)
{
	if(numMaps>=MAX_FUNCS)
	{
		printf("too many maps\n");
		getch();
	}
	maps[numMaps].name=strdup(a);
	maps[numMaps].dest=strdup(b);
	numMaps++;
}

char *GetFn(byte a,byte b,byte c)
{
	int32		i;
	char		tmp[128];

#ifdef EMIT_ASM
	if(a==0xDD)
	{
		if(b==0xCB)	sprintf(tmp,"Emulate_%02x_%02x_%02x:",a,b,c);
		else		sprintf(tmp,"Emulate_%02x_%02x:",a,b);
	}
	else if(a==0xFD)
	{
		if(b==0xCB)	sprintf(tmp,"Emulate_%02x_%02x_%02x:",a,b,c);
		else		sprintf(tmp,"Emulate_%02x_%02x:",a,b);
	}
	else if(a==0xCB)
	{
		sprintf(tmp,"Emulate_%02x_%02x:",a,b);
	}
	else if(a==0xED)
	{
		sprintf(tmp,"Emulate_%02x_%02x:",a,b);
	}
	else
	{
		sprintf(tmp,"Emulate_%02x:",a);
	}
#else
	if(a==0xDD)
	{
		if(b==0xCB)	sprintf(tmp,"void Emulate_%02x_%02x_%02x(void) {",a,b,c);
		else		sprintf(tmp,"void Emulate_%02x_%02x(void) {",a,b);
	}
	else if(a==0xFD)
	{
		if(b==0xCB)	sprintf(tmp,"void Emulate_%02x_%02x_%02x(void) {",a,b,c);
		else		sprintf(tmp,"void Emulate_%02x_%02x(void) {",a,b);
	}
	else if(a==0xCB)
	{
		sprintf(tmp,"void Emulate_%02x_%02x(void) {",a,b);
	}
	else if(a==0xED)
	{
		sprintf(tmp,"void Emulate_%02x_%02x(void) {",a,b);
	}
	else
	{
		sprintf(tmp,"void Emulate_%02x(void) {",a);
	}
#endif

	for(i=0;i<numMaps;i++)
	{
		if(!stricmp(maps[i].name,tmp))
		{
			return(maps[i].dest);
		}
	}
	return(NULL);
}

void Crunch(char *ii,char *oo)
{
	FILE		*in,*out;
	char		tmp[256];
	int32		st;
	int32		i,j,k;

	funcCt=-1;

	in=fopen(ii,"r");
	out=fopen(oo,"w");
	if(!in || !out)
	{
		exit(1);
	}

#ifdef EMIT_ASM
//	fprintf(out,".include \"z80core.h\"\n");
#else
	fprintf(out,"#include \"pc.h\"\n");
#endif
	st=-1;
	while(1)
	{
		fgets(tmp,sizeof(tmp),in);
		if(feof(in))	break;
		if(tmp[strlen(tmp)-1]=='\n')	tmp[strlen(tmp)-1]=0;
#ifdef EMIT_ASM
		if(tmp[0]=='@')
#else
		if(tmp[0]=='/')
#endif
		{
			funcCt++;
			funcs[funcCt]=new CFunc();
			funcs[funcCt]->com=strdup(tmp);
			st=1;
		}
		else if(st==1)
		{
			funcs[funcCt]->name=strdup(tmp);
			st++;
		}
		else
		{
			funcs[funcCt]->lines[funcs[funcCt]->lineCt]=strdup(tmp);
			funcs[funcCt]->lineCt++;
		}
	}
	if(st>=1)	funcCt++;
	fclose(in);

	printf("read %ld functions\n",funcCt);

	// compare functions, remove duplicates
	for(i=0;i<funcCt;i++)
	{
		for(j=i+1;j<funcCt;j++)
		{
			if(funcs[i]->lineCt != funcs[j]->lineCt)	continue;
			for(k=0;k<funcs[i]->lineCt;k++)
			{
				if(strcmp(funcs[i]->lines[k],funcs[j]->lines[k]))	break;
			}
			if(k!=funcs[i]->lineCt)
			{
				continue;
			}

			// identical!
			printf("%s is the same as %s\n",funcs[i]->name,funcs[j]->name);
			AddMapping(funcs[j]->name,funcs[i]->name);

			// remove duplicate
			memcpy(&funcs[j],&funcs[j+1],(funcCt-j)*sizeof(CFunc*));


			funcCt--;
			j--;
		}
	}

	for(i=0;i<funcCt;i++)
	{
		AddMapping(funcs[i]->name,funcs[i]->name);
	}

	printf("%ld unique functions remaining\n",funcCt);

	fprintf(out,"@\n");
	fprintf(out,"@ Z80CORE.S autogenerated by z80gen.exe\n");
	fprintf(out,"@ do not modify by hand!\n");
	fprintf(out,"@\n");

	for(i=0;i<funcCt;i++)
	{
		fprintf(out,"%s\n",funcs[i]->com);
		fprintf(out,"%s\n",funcs[i]->name);
		for(j=0;j<funcs[i]->lineCt;j++)
		{
			fprintf(out,"%s\n",funcs[i]->lines[j]);
		}
	}
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

/***********************************************************************
 *
 *
 *
 ***********************************************************************
 *
 *
 *
 ***********************************************************************/

int32		lineCt;
char		*lines[8192];
char		*ops[8192];
char		*args[8192];

void Test(void)
{
	FILE		*in,*out;
	int32		i;
	char	tmp[128];
	char	*c,*d;

	in=fopen("inst2.txt","rb");
	out=fopen("test.txt","wb");

	lineCt=0;
	while(1)
	{

		fgets(tmp,sizeof(tmp),in);
		if(feof(in)) break;
		if(tmp[strlen(tmp)-1]=='\n')	tmp[strlen(tmp)-1]=0;

		c=strstr(tmp,"\t");
		if(!c)	c=strstr(tmp," ");

		while((*c)=='\t' || (*c)==' ')	c++;
		d=strstr(c,"\t");
		if(d)	*d=0;

		d=strstr(c,"*");
		if(d)	*d=0;
		d=strstr(c,"\n");
		if(d)	*d=0;
		d=strstr(c,"\r");
		if(d)	*d=0;

//		fprintf(out,"%s\n",c);

		for(i=0;i<lineCt;i++)
		{
			if(!strcmp(lines[i],c))	break;
		}
		if(i==lineCt)
		{
			lines[lineCt++]=strdup(c);
		}
	}

	for(i=0;i<lineCt;i++)
	{
		c=strstr(lines[i]," ");
		if(c)
		{
			*c=0;
			ops[i]=strdup(lines[i]);
			args[i]=strdup(c+1);
		}
		else
		{
			ops[i]=strdup(lines[i]);
			args[i]="";
		}
	}

	printf("%ld unique\n",lineCt);

	for(i=0;i<lineCt;i++)
	{
		fprintf(out,"[%s] [%s]\n",ops[i],args[i]);
	}
}

